from . import (io, utils, decoding, localize, selectivity, spikes,
               spike_distance, stats, viz)

from .viz import plot_shaded, plot_raster, plot_spikes
from .spikes import SpikeEpochs, Spikes
from .selectivity import depth_of_selectivity, explained_variance


# -*- coding: utf-8 -*-
"""
Anonymize DICOM files
---------------------
Simple script to anonymize DICOM images for multiple patients.
The folders in `input_path` should be folders named as patients (for example
U07, U08, etc.) and each contains `ct` and/or `mri` subfolders with DICOM
images to anonymize.

Created on Fri Nov 11 01:37:51 2022

@author: mmagn
"""
import os
import dicomanonymizer as diano


input_path = r'G:\My Drive\syracuse_anat'
output_path = r'G:\My Drive\syracuse_anat_anon'


subdirs = os.listdir(input_path)
for subdir in subdirs:
    full_subdir = os.path.join(input_path, subdir)
    sub_subdirs = os.listdir(full_subdir)

    has_ct = 'ct' in sub_subdirs
    has_mri = 'mri' in sub_subdirs

    if has_ct or has_mri:
        # make sure subject folder is present in output path
        output_subdir = os.path.join(output_path, subdir)
        if not os.path.exists(output_subdir):
            os.mkdir(output_subdir)
        print(f'Processing patient {subdir}...')

    if has_ct:
        # anonymize CT scans
        print('Anonymizing CT scans')
        dicom_path_in = os.path.join(full_subdir, 'ct')
        dicom_path_out = os.path.join(output_subdir, 'ct')

        if not os.path.exists(dicom_path_out):
            os.mkdir(dicom_path_out)

        diano.anonymize(dicom_path_in, dicom_path_out, {}, False)

    if has_mri:
        # anonymize mri scans
        print('Anonymizing MRI scans')
        dicom_path_in = os.path.join(full_subdir, 'mri')
        dicom_path_out = os.path.join(output_subdir, 'mri')

        if not os.path.exists(dicom_path_out):
            os.mkdir(dicom_path_out)

        diano.anonymize(dicom_path_in, dicom_path_out, {}, False)



import numpy as np
from .utils import (_deal_with_picks, _turn_spike_rate_to_xarray,
                    _gauss_kernel_samples, _symmetric_window_samples)



# TODO: add n_jobs?
# CONSIDER wintype 'rectangular' vs 'gaussian'
def compute_spike_rate(spk, picks=None, winlen=0.25, step=0.01, tmin=None,
                       tmax=None, backend='numpy'):
    '''Calculate spike rate with a running or static window.

    Parameters
    ----------
    spk : pylabianca.SpikeEpochs
        Spikes object to compute firing rate for.
    picks : int | listlike of int | None
        The neuron indices / names to use in the calculations. The default
        (``None``) uses all cells.
    winlen : float
        Length of the running window in seconds.
    step : float | bool
        The step size of the running window. If step is ``False`` then
        spike rate is not calculated using a running window but with
        a static one with limits defined by ``tmin`` and ``tmax``.
    tmin : float | None
        Time start in seconds. Default to trial start if ``tmin`` is ``None``.
    tmax : float | None
        Time end in seconds. Default to trial end if ``tmax`` is ``None``.
    backend : str
        Execution backend. Can be ``'numpy'`` or ``'numba'``.

    Returns
    -------
    frate : xarray.DataArray
            Xarray with following labeled dimensions: cell, trial, time.
    '''
    picks = _deal_with_picks(spk, picks)
    tmin = spk.time_limits[0] if tmin is None else tmin
    tmax = spk.time_limits[1] if tmax is None else tmax

    if isinstance(step, bool) and not step:

        times = f'{tmin} - {tmax} s'

        frate = list()
        cell_names = list()

        for pick in picks:
            frt = _compute_spike_rate_fixed(
                spk.time[pick], spk.trial[pick], [tmin, tmax],
                spk.n_trials)
            frate.append(frt)
            cell_names.append(spk.cell_names[pick])

    else:
        frate = list()
        cell_names = list()

        if backend == 'numpy':
            func = _compute_spike_rate_numpy
        elif backend == 'numba':
            from ._numba import _compute_spike_rate_numba
            func = _compute_spike_rate_numba
        else:
            raise ValueError('Backend can be only "numpy" or "numba".')

        for pick in picks:
            times, frt = func(
                spk.time[pick], spk.trial[pick], np.array([tmin, tmax]),
                spk.n_trials, winlen=winlen, step=step)
            frate.append(frt)
            cell_names.append(spk.cell_names[pick])

    frate = np.stack(frate, axis=0)
    frate = _turn_spike_rate_to_xarray(times, frate, spk,
                                        cell_names=cell_names)
    frate = _add_frate_info(frate)

    return frate


def _add_frate_info(arr, dep='rate'):
    arr.name = f'firing {dep}'
    arr.attrs['unit'] = "Hz"
    arr.attrs['coord_units'] = {'time': 's'}
    return arr


# ENH: speed up by using previous mask in the next step to pre-select spikes
def _compute_spike_rate_numpy(spike_times, spike_trials, time_limits,
                              n_trials, winlen=0.25, step=0.05):
    half_win = winlen / 2
    window_limits = np.array([-half_win, half_win])
    used_range = time_limits[1] - time_limits[0]
    n_steps = int(np.floor((used_range - winlen) / step + 1))

    fr_t_start = time_limits[0] + half_win
    fr_t_end = time_limits[1] - half_win + step * 0.001
    times = np.arange(fr_t_start, fr_t_end, step=step)
    frate = np.zeros((n_trials, n_steps))

    for step_idx in range(n_steps):
        win_lims = times[step_idx] + window_limits
        msk = (spike_times >= win_lims[0]) & (spike_times < win_lims[1])
        tri = spike_trials[msk]
        in_tri, count = np.unique(tri, return_counts=True)
        frate[in_tri, step_idx] = count / winlen

    return times, frate


# @numba.jit(nopython=True)
# currently raises warnings, and jit is likely not necessary here
# Encountered the use of a type that is scheduled for deprecation: type
# 'reflected list' found for argument 'time_limits' of function
# '_compute_spike_rate_fixed'
def _compute_spike_rate_fixed(spike_times, spike_trials, time_limits,
                              n_trials):

    winlen = time_limits[1] - time_limits[0]
    frate = np.zeros(n_trials)
    msk = (spike_times >= time_limits[0]) & (spike_times < time_limits[1])
    tri = spike_trials[msk]
    in_tri, count = np.unique(tri, return_counts=True)
    frate[in_tri] = count / winlen

    return frate


# TODO: add n_jobs?
# TODO: consider an exact mode where the spikes are not transformed to raw /
#       binned representation (binned raster) and the gaussian kernel is placed
#       exactly where the spike is (`loc=spike_time`?) and evaluated
#       (maybe this is what is done by elephant?)
# TODO: check if time is symmetric wrt 0 (in most cases it should be as epochs
#       are constructed wrt specific event)
def _spike_density(spk, picks=None, winlen=0.3, gauss_sd=None, fwhm=None,
                   kernel=None, sfreq=500.):
    '''Calculates normal (constant) spike density.

    The density is computed by convolving the binary spike representation
    with a gaussian kernel.

    Parameters
    ----------
    spk : SpikeEpochs
        SpikeEpochs object.
    picks : array-like | None
        Indices or names of cells to use. If ``None`` all cells are used.
    winlen : float
        Length of the gaussian kernel in seconds. Default is ``0.3``.
        If ``gauss_sd`` is ``None`` the standard deviation of the gaussian
        kernel is set to ``winlen / 6``.
    gauss_sd : float | None
        Standard deviation of the gaussian kernel in seconds. If ``None``
        the standard deviation is set to ``winlen / 6``.
    fwhm : float | None
        Full width at half maximum of the gaussian kernel in seconds.
    kernel : array-like | None
        Kernel to use for convolution. If ``None`` the gaussian kernel is
        constructed from ``winlen`` and ``gauss_sd``.
    sfreq : float
        Sampling frequency (in Hz) of the spike density representation. Default
        is ``500``.
    '''
    from scipy.signal import correlate

    if kernel is None:
        if fwhm is not None:
            gauss_sd = _gauss_sd_from_FWHM(fwhm)
            winlen = gauss_sd * 6.
        else:
            gauss_sd = winlen / 6. if gauss_sd is None else gauss_sd

        gauss_sd = gauss_sd * sfreq
        win_smp, trim = _symmetric_window_samples(winlen, sfreq)
        kernel = _gauss_kernel_samples(win_smp, gauss_sd) * sfreq
    else:
        assert (len(kernel) % 2) == 1
        trim = int((len(kernel) - 1) / 2)

    picks = _deal_with_picks(spk, picks)
    times, binrep = spk.to_raw(picks=picks, sfreq=sfreq)
    cnt_times = times[trim:-trim]

    cnt = correlate(binrep, kernel[None, None, :], mode='valid')
    return cnt_times, cnt


def _gauss_sd_from_FWHM(FWHM):
    """Calculate the standard deviation of a gaussian kernel from the FWHM.

    Parameters
    ----------
    FWHM : float
        Full width at half maximum of the gaussian kernel in seconds.

    Returns
    -------
    gauss_sd : float
        Standard deviation of the gaussian kernel in seconds.
    """
    gauss_sd = FWHM / (2 * np.sqrt(2 * np.log(2)))
    return gauss_sd


def _FWHM_from_window(winlen=None, gauss_sd=None):
    """Calculate the full width at half maximum of a gaussian kernel.

    Parameters
    ----------
    winlen : float | None
        Length of the gaussian kernel in seconds. The length is assumed to be
        6 times the standard deviation of the gaussian kernel. Exactly one of
        ``winlen`` and ``gauss_sd`` must be specified. Default is ``None``,
        in which case ''winlen`` is ignored and ``gauss_sd`` is used.
    gauss_sd : float | None
        Standard deviation of the gaussian kernel in seconds. Exactly one of
        ``winlen`` and ``gauss_sd`` must be specified. Default is ``None``,
        in which case ''gauss_sd`` is ignored and ``winlen`` is used.

    Returns
    -------
    FWHM : float
        Full width at half maximum of the gaussian kernel in seconds.
    """
    # exactly one of the two must be specified
    assert winlen is not None or gauss_sd is not None
    assert winlen is None or gauss_sd is None

    gauss_sd = winlen / 6 if gauss_sd is None else gauss_sd
    FWHM = 2 * np.sqrt(2 * np.log(2)) * gauss_sd
    return FWHM



import warnings
import numpy as np

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.utils.multiclass import unique_labels


def run_decoding_array(X, y, n_splits=6, C=1., scoring='accuracy',
                       n_jobs=1, time_generalization=False, random_state=None,
                       clf=None, n_pca=0, feature_selection=None):
    '''Perform decoding analysis.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features, n_times)
        The training input samples. Commonly the samples dimension consists of
        trials or subjects while the features dimension consists of cells or
        electrodes.
    y : array-like, shape (n_samples,)
        The target values (class labels).
    n_splits : int | str
        Number of cross-validation splits. If ``'loo'``, leave-one-out
        cross-validation is used.
    C : float
        Inverse of regularization strength.
    scoring : str
        Scoring metric.
    n_jobs : int
        Number of jobs to run in parallel.
    time_generalization : bool
        Whether to perform time generalization (training and testing also
        on different time points).
    random_state : int or None
        Random state for cross-validation.
    clf : None or sklearn classifier / pipeline
        If None, a linear SVM classifier with standard scaling is used.
    n_pca : int
        Number of principal components to use for dimensionality reduction. If
        0 (default), no dimensionality reduction is performed.
    feature_selection : function | None
        Function that takes ``X`` and ``y`` array from training set and
        returns a boolean array of shape (n_features,) that indicates which
        features to use for training. If None (default), no feature selection
        is performed.

    Returns
    -------
    scores : array, shape (n_splits, n_times, n_times)
        Decoding scores.
    '''
    from sklearn.pipeline import make_pipeline
    from sklearn.preprocessing import StandardScaler
    from sklearn.svm import SVC
    from sklearn.model_selection import StratifiedKFold, LeaveOneOut
    from mne.decoding import GeneralizingEstimator, SlidingEstimator

    if n_pca > 0:
        if clf is not None:
            raise ValueError('Cannot use PCA and a custom classifier.'
                             ' You would have to construct you own pipeline.')

        from sklearn.decomposition import PCA
        pca = PCA(n_components=n_pca)

    # handle data with only one time point / aggregated time window
    one_time_sample = False
    if X.ndim == 2:
        one_time_sample = True

    # k-fold object
    if isinstance(n_splits, str) and n_splits == 'loo':
        # use leave one out cross validation
        spl = LeaveOneOut()
    else:
        spl = StratifiedKFold(
            n_splits=n_splits,
            shuffle=True,
            random_state=random_state
        )

    # classification pipeline
    if clf is None:
        steps = [StandardScaler(), SVC(C=C, kernel='linear')]
        if n_pca > 0:
            steps.insert(1, pca)
        clf = make_pipeline(*steps)

    # use simple sliding estimator or generalization across time
    if not one_time_sample:
        estimator = (SlidingEstimator if not time_generalization
                     else GeneralizingEstimator)
        estimator = estimator(
            clf, scoring=scoring,
            n_jobs=n_jobs, verbose=False
        )
    else:
        estimator = clf

    # do the k-fold
    scores = list()
    for train_index, test_index in spl.split(X, y):
        if feature_selection is not None:
            sel = feature_selection(X[train_index, :], y[train_index])
        else:
            sel = slice(None)

        estimator.fit(X=X[train_index][:, sel],
                      y=y[train_index])
        score = estimator.score(X=X[test_index][:, sel],
                                y=y[test_index])
        scores.append(score)

    scores = np.stack(scores, axis=0)
    return scores


# CONSIDER: decim=None by default, decim=1 as no decimation may be confusing
# CONSIDER: supporting ``select`` to select conditions (useful only when a
#           dictionary of xarrays is passed, so multiple subjects)
def run_decoding(arr, target, decode_across='time', decim=1, n_splits=6, C=1.,
                 scoring='accuracy', n_jobs=1, time_generalization=False,
                 random_state=None, clf=None, n_pca=0, feature_selection=None):
    '''Perform decoding analysis using xarray as input.

    Parameters
    ----------
    arr : xarray.DataArray
        The data to use in classification. Should contain a ``'trials'``
        dimension and a dimension consistent with ``decode_across``.
    target : str
        Name of the variable to use as target.
    decode_across : str
        Name of the dimension to perform decoding across. For each element of
        this dimension a separate decoding will be performed. The default is
        ``'time'`` - which slides the decoding classifier across the time
        dimension.
    decim : int
        Decimation for the ``decode_across`` dimension. Default's to ``1``,
        which does not perform decimation.
    n_splits : int | str
        Number of cross-validation splits. If ``'loo'``, leave-one-out
        cross-validation is used.
    C : float
        Inverse of regularization strength.
    scoring : str
        Scoring metric.
    n_jobs : int
        Number of jobs to run in parallel.
    time_generalization : bool
        Whether to perform time generalization (training and testing also
        on different time points).
    random_state : int or None
        Random state for cross-validation.
    clf : None or sklearn classifier / pipeline
        If None, a linear SVM classifier with standard scaling is used.
    n_pca : int
        Number of principal components to use for dimensionality reduction. If
        0 (default), no dimensionality reduction is performed.
    feature_selection : function | None
        Function that takes ``X`` and ``y`` array from training set and
        returns a boolean array of shape (n_features,) that indicates which
        features to use for training. If None (default), no feature selection
        is performed.

    Returns
    -------
    scores : xarray
        Decoding scores. The first dimension should consist of folds (separate
        cross-validation folds).
    '''
    import xarray as xr
    assert isinstance(arr, xr.DataArray)
    assert decode_across in arr.dims

    orig_dims = list(arr.dims)

    X, y, time_dim = frate_to_sklearn(
        arr, target=target, select=None, decim=decim)
    scores = run_decoding_array(
        X, y, n_splits=n_splits, C=C, scoring=scoring, n_jobs=n_jobs,
        time_generalization=time_generalization, random_state=random_state,
        clf=clf, n_pca=n_pca, feature_selection=feature_selection
    )

    # combine into xarray output
    name = scoring
    coords = {'fold': np.arange(n_splits)}
    if time_generalization:
        dims = ['fold', 'train_' + decode_across, 'test_' + decode_across]
        coords[dims[1]] = time_dim
        coords[dims[2]] = time_dim
    else:
        dims = ['fold'] + [decode_across]
        coords[decode_across] = time_dim

    scores = xr.DataArray(
        scores, dims=dims, coords=coords, name=name,
    )
    return scores


# TODO: later may be useful to make it accept arrays with different dimension
#       names
def frate_to_sklearn(frate, target=None, select=None,
                     cell_names=None, time_idx=None, decim=10):
    '''Formats xarray.DataArray into sklearn X and y data arrays.

    Parameters
    ----------
    frate : xarray.DataArray
        The data to turn in sklearn X and y format. Should contain a
        ``'trials'``, ``'cell'`` and ``'time'`` dimension.
    target : str
        Name of the variable to use as target.
    select : str | None
        Condition query to subselect trials. If ``None`` no trial selection
        is performed (all trials are used).
    cell_names : list of str | None
        Cell names to select.
    time_idx : int | array of int | None
        Select specific time point or timepoints. If ``None``, no specific
        timepoints are picked (apart from decimation controlled by ``decim``).
    decim : int
        Decimation factor.

    Returns
    -------
    X : numpy.ndarray
        2D array of n_observations x n_features (x n_times) shape. The time
        dimension is added only when it is present in the input xarray.
    y : numpy.ndarray
        1D array of n_observations length with target categories to classify.
    time : numpy.ndarray
        1D array of time labels after decimation.
    '''
    if target is None:
        raise ValueError('You have to specify specify target.')

    has_time = 'time' in frate.dims
    if has_time:
        fr = frate.transpose('trial', 'cell', 'time')
    else:
        fr = frate.transpose('trial', 'cell')

    # if concat_cond is not None:
    #     fr2 = frates['maint1'][subj].transpose(
    #         'trial', 'cell', 'time').sel(time=slice(-0.25, 2.5))

    if cell_names is not None:
        fr = fr.sel(cell=cell_names)

    if select is not None:
        fr = fr.query({'trial': select})

    if time_idx is not None:
        fr = fr.isel(time=time_idx)

    if has_time:
        time = fr.time.values[::decim]
        X = fr.values[..., ::decim]
    else:
        time = None
        X = fr.values

    y = fr.coords[target].values

    return X, y, time


def frates_dict_to_sklearn(frates, target=None, select=None,
                           cell_names=None, time_idx=None, decim=10):
    '''Get all subjects from frates dictionary.

    The ``frates`` is a dictionary of subject -> firing rate xarray mappings.

    Parameters
    ----------
    frates : dict
        Dictionary of the form {subject_string: firing rate xarray}.
    target : str
        Name of the variable to use as target.
    select : str, optional
        Query string to select trials. Default is ``None``, which does not
        subselect trials (all trials are used).
    cell_names : dict of list of str, optional
        Dictionary with subject id as keys and values being list of cell names
        to use. Defaults to ``None``, which uses all cells.
    time_idx : int, optional
        Time index or time range to use (as time indices). Defaults to
        ``None``, which uses all time points.
    decim : int
        Decimation factor.

    Returns
    -------
    Xs : list of arrays
        List of (n_trials, n_cells, n_time) firing rate arrays extracted from
        the ``frates`` dictionary. Each list element represents one subject.
    ys : list of arrays
        List of (n_trials,) target values extracted from the ``frates``
        dictionary. Each list element represents one subject.
    full_time : np.array | None
        Full time vector (in seconds). ``None`` if no time dimension was
        present in the the xarrays in ``frates`` dictionary.
    '''
    Xs = list()
    ys = list()

    if target is None:
        raise ValueError('You have to specify specify target.')

    if cell_names is None:
        subjects = frates.keys()
    else:
        subjects = cell_names.keys()

    for subj in subjects:
        this_cell_names = (cell_names[subj] if cell_names is not None
                           else cell_names)
        X, this_y, full_time = frate_to_sklearn(
            frates[subj], select=select, target=target,
            cell_names=this_cell_names, time_idx=time_idx, decim=decim
        )

        # add to the list
        Xs.append(X)
        ys.append(this_y)

    return Xs, ys, full_time


def join_subjects(Xs, ys, random_state=None, shuffle=True):
    '''Concatenate subjects keeping target the same but shuffling tirals
    within target categories.

    The concatenated array is trials x subjects_cells x time.

    Parameters
    ----------
    Xs : list of arrays
        List of (n_trials, n_cells, n_time) firing rate arrays.
    ys : list of arrays
        List of (n_trials,) target values.
    random_state : int, optional
        Random state to use for shuffling within-condition trials per subject
        before joining the arrays into one "pseudo-population".
    shuffle : bool, optional
        Whether to shuffle trials within conditions before joining the arrays.

    Returns
    -------
    X : array
        (n_trials, n_cells, n_time) firing rate array. The cells dimension
        is a pseudo-population pooled from all the subjects.
    y : array
        (n_trials,) target values.
    '''
    n_arrays = len(Xs)
    new_Xs, new_ys = list(), list()

    if random_state is not None:
        rnd = np.random.default_rng(random_state)
    else:
        rnd = None

    for idx in range(n_arrays):

        # shuffle
        X, y = Xs[idx], ys[idx]

        if shuffle:
            X, y = shuffle_trials(X, y, random_state=rnd)

            # sort trials by dependent value
            srt = np.argsort(y)
        else:
            # normal sorting can change the order of trials within conditions
            # so here, we do simpler reordering that retains this order
            categories = np.unique(y)
            srt = np.concatenate([np.where(y == cat)[0]
                                  for cat in categories])

        X = X[srt, :]
        y = y[srt]

        # add to the list
        new_Xs.append(X)
        new_ys.append(y)

    # make sure trials are aligned
    assert all([(y == new_ys[0]).all() for y in new_ys])

    X = np.concatenate(new_Xs, axis=1)
    y = new_ys[0]

    return X, y


# CONSIDER: to not have to copy arr, we can allow `target` to get a vector of
#           values
def permute(arr, decoding_fun, target=None, n_permutations=200, n_jobs=1,
            average_folds=True, arguments=dict()):
    import pandas as pd
    import xarray as xr

    scores = list()
    arr = arr.copy()  # we copy to modify target
    target_coord = arr.coords[target].values

    if 'target' not in arguments:
        arguments['target'] = target

    if n_jobs > 1:
        from joblib import Parallel, delayed

        # perocess in parallel only at the top (permutation) level
        arguments['n_jobs'] = 1

        scores = Parallel(n_jobs=n_jobs)(
            delayed(_do_permute)(
                arr, decoding_fun, target_coord,
                average_folds=average_folds,
                arguments=arguments
            )
            for perm_idx in range(n_permutations)
        )
    else:
        for _ in range(n_permutations):
            score = _do_permute(
                arr, decoding_fun, target_coord, average_folds=average_folds,
                arguments=arguments)
            scores.append(score)

    # join the results
    if isinstance(scores[0], xr.DataArray):
        perm = pd.Index(np.arange(n_permutations), name='permutation')
        scores = xr.concat(scores, perm)
    else:
        scores = np.stack(scores, axis=0)
    return scores


def _do_permute(arr, decoding_fun, target_coord, average_folds=True,
                arguments=dict()):
    # permute target
    np.random.shuffle(target_coord)

    scr = decoding_fun(arr, **arguments)

    if average_folds:
        scr = scr.mean(dim='fold')

    return scr


def resample_decoding(decoding_fun, frates=None, target=None, Xs=None, ys=None,
                      time=None, arguments=dict(), n_resamples=20, n_jobs=1,
                      permute=False, select_trials=None, decim=None):
    """Resample a decoding analysis.

    The resampling is done by rearranging trials within each subject,
    matching the ``target`` category across subjects.

    Parameters
    ----------
    decoding_fun : callable
        Decoding function to use. Must accept ``X`` and ``y`` as first two
        arguments, and allow for ``time`` keyword argument. When given the
        ``time`` argument, the function must return a xarray DataArray with
        decoding scores.
    frates : dict | None
        Dictionary of the form {subject_string: firing rate xarray}. If
        ``None``, the ``Xs`` and ``ys`` arguments must be provided.
    target : str | None
        Target category to use for decoding. Has to be provided if ``frates``
        is used.
    Xs : list of arrays | None
        List of (n_trials, n_cells, n_time) firing rate arrays (one array per
        subject / session). If ``None``, the ``frates`` argument must be
        provided.
    ys : list of arrays | None
        List of (n_trials,) target values (one array per subject / session).
        If ``None``, the ``frates`` argument must be provided.
    time : array | None
        If data is provided as ``Xs`` and ``ys`` arrays, a time vector is
        needed.
    arguments : dict, optional
        Additional arguments to pass to the ``decoding_fun``.
    n_resamples : int, optional
        Number of resamples to perform. Defaults to 20.
    n_jobs : int, optional
        Number of jobs to use for resampling. Defaults to 1, higher values
        will use joblib to parallelize the resampling.
    permute : bool | array, optional
        Whether to permute the trials within each subject. If ``True``, the
        trials are permuted randomly. If an array, the array is used to
        permute the trials. Defaults to ``False``, which does not perform
        permutation.
    select_trials : str, optional
        A string query used to select trials out of subject-specific xarrays.
        Defaults to ``None``, which uses all trials.
    decim : int | None, optional
        Temporal decimation factor. Defaults to ``None``, which does not
        perform decimation.

    Returns
    -------
    score_resamples : xarray.DataArray | numpy.ndarray
        Resampled decoding scores. If the ``decoding_fun`` returns a xarray
        DataArray, the resampled scores are also returned as a DataArray.
    """
    import xarray as xr

    # check if correct arguments are provided
    if frates is None:
        if Xs is None or ys is None:
            raise ValueError('Either frates or Xs and ys must be provided.')
    else:
        if target is None:
            raise ValueError('target must be provided if frates is used.')

        Xs, ys, time = frates_dict_to_sklearn(
            frates, target=target, select=select_trials, decim=decim)

    n_trials = _count_trials(Xs)

    if isinstance(permute, bool) and permute:
        permute = np.arange(n_trials)
        np.random.shuffle(permute)

    # split into n_jobs = 1 or n_resamples = 1
    # and n_jobs > 1 (joblib)
    if n_jobs == 1 or n_resamples == 1:
        score_resamples = [
            _do_resample(Xs, ys, decoding_fun, arguments,
                         permute=permute, time=time)
            for resample_idx in range(n_resamples)
        ]
    else:
        from joblib import Parallel, delayed

        score_resamples = Parallel(n_jobs=n_jobs)(
            delayed(_do_resample)(
                Xs, ys, decoding_fun, arguments, permute=permute, time=time)
            for resample_idx in range(n_resamples)
        )

    # join the results
    if isinstance(score_resamples[0], xr.DataArray):
        import pandas as pd
        resamples = pd.Index(np.arange(n_resamples), name='resample')
        score_resamples = xr.concat(score_resamples, resamples)
    else:
        score_resamples = np.stack(score_resamples, axis=0)
    return score_resamples


def _do_resample(Xs, ys, decoding_fun, arguments, permute=False, time=None):
    X, y = join_subjects(Xs, ys)

    if isinstance(permute, bool):
        if permute:
            np.random.shuffle(y)
    else:
        # assume permutation array
        y = y[permute]

    # do the actual decoding
    return decoding_fun(X, y, time=time, **arguments)


def _count_trials(Xs):
    '''Check trials foe each array in the list.'''
    # check n trials (across subjects)

    n_tri = np.array([X.shape[0] for X in Xs])
    assert (n_tri[0] == n_tri).all()
    n_tri = n_tri[0]

    return n_tri


def shuffle_trials(*arrays, random_state=None):
    '''Perform the same trial shuffling for multiple arrays.'''
    n_arrays = len(arrays)
    array_len = [len(x) for x in arrays]
    if n_arrays > 1:
        assert all([array_len[0] == x for x in array_len[1:]])

    n_trials = array_len[0]
    tri_idx = np.arange(n_trials)
    if random_state is None:
        np.random.shuffle(tri_idx)
    else:
        random_state.shuffle(tri_idx)

    shuffled = tuple(arr[tri_idx] for arr in arrays)
    if n_arrays == 1:
        shuffled = shuffled[0]
    return shuffled


def select_n_best_cells(X, y, select_n=1):
    '''Select n best cells (used on training data).

    Parameters
    ----------
    X : numpy.ndarray
        sklearn X array: ``n_observations x n_features (x n_samples)``.
    y : numpy.ndarray
        Vector of class id to predict.
    select_n : int
        The number of best-performing units to select.
    '''
    from scipy.stats import ttest_ind

    t_val_per_cell, _ = np.abs(ttest_ind(X[y], X[~y]))
    if t_val_per_cell.ndim > 1:
        # max over time
        t_val_per_cell = t_val_per_cell.max(axis=1)
    tval_ord = t_val_per_cell.argsort()[::-1]

    # return selector array
    n_features = t_val_per_cell.shape[0]
    sel = np.zeros(n_features, dtype='bool')
    sel[tval_ord[:select_n]] = True
    return sel


def correlation(X1, X2):
    '''Correlate two arrays.'''
    ncols1 = X1.shape[1]
    rval = np.corrcoef(X1, X2, rowvar=False)
    rval_sel = rval[:ncols1, ncols1:]
    return rval_sel


class maxCorrClassifier(BaseEstimator):
    '''Simple implementation of maxCorr classifier.'''
    def __init__(self):
        '''Create an instance of maxCorr classifier.'''
        pass

    def fit(self, X, y, scoring=None):
        '''Fit maxCorr to training data.

        Parameters
        ----------
        X : np.array
            Training data to use in classification. n_observations x n_features
            numpy array.
        y : np.array
            Target classes to classify. n_observations vector.
        scoring : str | None
            Scikit-learn scoring method.

        Returns
        -------
        self : pylabianca.decoding.maxCorr
            Fit classifier to training data. Works in-place so the output
            does not have to be stored (but it's useful for chaining).
        '''
        X, y = check_X_y(X, y)
        self.classes_ = unique_labels(y)
        self.class_averages_ = list()
        self.n_classes_ = len(self.classes_)
        self.n_features_in_ = X.shape[1]
        for cls in self.classes_:
            msk = y == cls
            avg = X[msk, :].mean(axis=0)
            self.class_averages_.append(avg)

        self.class_averages_ = np.stack(self.class_averages_, axis=1)
        self.scoring = 'accuracy' if scoring is None else scoring

        return self

    def predict(self, X):
        '''
        Predict classes.

        Parameters
        ----------
        X : numpy.array
            Test data to use in predicting classes. Should be an
            n_observations x n_features numpy array.

        Returns
        -------
        y_pred : numpy.array
            Vector of predicted classes.
        '''
        # Check if fit has been called
        check_is_fitted(self)
        X = check_array(X)

        # check if any of the classes is constant
        has_constant_class = (
            self.class_averages_ == self.class_averages_[[0], :]
            ).all(axis=0).any()

        if not has_constant_class:
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    'ignore',
                    message='invalid value encountered in true_divide',
                    category=RuntimeWarning
                )
                r = correlation(self.class_averages_, X.T)
        else:
            distance = self.class_averages_[..., None] - X.T[:, None, :]
            r = np.linalg.norm(distance, axis=0) * -1

        bad_trials = np.isnan(r).all(axis=0)
        if bad_trials.any():
            distance = (self.class_averages_[..., None]
                        - X.T[:, None, bad_trials])
            r[:, bad_trials] = np.linalg.norm(distance, axis=0) * -1

        # pick class with best correlation:
        r_best = r.argmax(axis=0)
        y_pred = self.classes_[r_best]

        return y_pred

    def score(self, X=None, Y=None):
        '''
        Score classification.

        The scoring provided at initialization is used.

        Parameters
        ----------
        X : numpy.array
            Test data to use in predicting classes. Should be an
            n_observations x n_features numpy array.
        y : numpy.array
            Correct class labels. n_observations vector.
        '''
        from sklearn.metrics import get_scorer

        scorer = get_scorer(self.scoring)
        score = scorer(self, X, Y)
        return score


class SparseRemover(BaseEstimator, TransformerMixin):
    def __init__(self, n_select, threshold=1e-5):
        self.n_select = n_select
        self.threshold = threshold

    def fit(self, X, y):
        X, y = check_X_y(X, y)

        bad_prop = (X < self.threshold).mean(axis=0)
        bad_prop_thresh = np.sort(bad_prop)[self.n_select]
        self.sel_features_ = bad_prop <= bad_prop_thresh

        return self

    def transform(self, X):
        # Check if fit has been called
        check_is_fitted(self)
        X = check_array(X)

        return X[:, self.sel_features_]

    def fit_transform(self, X, y):
        self.fit(X, y)
        return self.transform(X)


class SelectKBestLeastSparse(BaseEstimator, TransformerMixin):
    def __init__(self, stat_fun, k=10, threshold=1e-5):
        self.n_select = k
        self.stat_fun = stat_fun
        self.threshold = threshold

    def fit(self, X, y):
        X, y = check_X_y(X, y)

        bad_prop = (X < self.threshold).mean(axis=0)
        stat, _ = self.stat_fun(X, y)
        select = stat * (1 - bad_prop)
        self.sel_features_ = np.argsort(select)[::-1][:self.n_select]

        return self

    def transform(self, X):
        # Check if fit has been called
        check_is_fitted(self)
        X = check_array(X)

        return X[:, self.sel_features_]

    def fit_transform(self, X, y):
        self.fit(X, y)
        return self.transform(X)



# coding=utf-8

"""
The MIT License (MIT)

Copyright (c) 2015 Aleksander Alafuzoff

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import division

import os
import warnings
import numpy as np
import datetime

HEADER_LENGTH = 16 * 1024  # 16 kilobytes of header

NCS_SAMPLES_PER_RECORD = 512
NCS_RECORD = np.dtype([('TimeStamp',       np.uint64),       # Cheetah timestamp for this record. This corresponds to
                                                             # the sample time for the first data point in the Samples
                                                             # array. This value is in microseconds.
                       ('ChannelNumber',   np.uint32),       # The channel number for this record. This is NOT the A/D
                                                             # channel number
                       ('SampleFreq',      np.uint32),       # The sampling frequency (Hz) for the data stored in the
                                                             # Samples Field in this record
                       ('NumValidSamples', np.uint32),       # Number of values in Samples containing valid data
                       ('Samples',         np.int16, NCS_SAMPLES_PER_RECORD)])  # Data points for this record. Cheetah
                                                                                # currently supports 512 data points per
                                                                                # record. At this time, the Samples
                                                                                # array is a [512] array.

NEV_RECORD = np.dtype([('stx',           np.int16),      # Reserved
                       ('pkt_id',        np.int16),      # ID for the originating system of this packet
                       ('pkt_data_size', np.int16),      # This value should always be two (2)
                       ('TimeStamp',     np.uint64),     # Cheetah timestamp for this record. This value is in
                                                         # microseconds.
                       ('event_id',      np.int16),      # ID value for this event
                       ('ttl',           np.int16),      # Decimal TTL value read from the TTL input port
                       ('crc',           np.int16),      # Record CRC check from Cheetah. Not used in consumer
                                                         # applications.
                       ('dummy1',        np.int16),      # Reserved
                       ('dummy2',        np.int16),      # Reserved
                       ('Extra',         np.int32, 8),   # Extra bit values for this event. This array has a fixed
                                                         # length of eight (8)
                       ('EventString',   'S', 128)])  # Event string associated with this event record. This string
                                                         # consists of 127 characters plus the required null termination
                                                         # character. If the string is less than 127 characters, the
                                                         # remainder of the characters will be null.

VOLT_SCALING = (1, u'V')
MILLIVOLT_SCALING = (1000, u'mV')
MICROVOLT_SCALING = (1000000, u'µV')


def read_header(file_path):
    '''Reads and parses the header of a Neuralynx file.

    Parameters
    ----------
    filename : str
        Path to the file.

    Returns
    -------
    header : dict
        Dictionary representing the header.
    '''
    file_path = os.path.abspath(file_path)
    with open(file_path, 'rb') as fid:
        raw_header = read_raw_header(fid)

    header = parse_header(raw_header)
    return header


def read_raw_header(fid):
    # Read the raw header data (16 kb) from the file object fid. Restores the position in the file object after reading.
    pos = fid.tell()
    fid.seek(0)
    raw_hdr = fid.read(HEADER_LENGTH).strip(b'\0')
    fid.seek(pos)

    return raw_hdr


def _get_field_value(hdr_lines, field_name):
    line_idx = [idx for idx, txt in enumerate(hdr_lines)
                if txt.startswith(field_name)]

    if len(line_idx) == 0:
        warnings.warn(f'Could not find {field_name} in Neuralynx header.')
        idx = None
        value = None
    else:
        idx = line_idx[0]
        value = ' '.join(hdr_lines[idx].split(' ')[1:])

    return idx, value


def parse_header(raw_hdr):
    # Parse the header string into a dictionary of name value pairs
    hdr = dict()

    # Decode the header as iso-8859-1 (the spec says ASCII, but there is at least one case of 0xB5 in some headers)
    raw_hdr = raw_hdr.decode('iso-8859-1')

    # Neuralynx headers seem to start with a line identifying the file, so
    # let's check for it
    hdr_lines = [line.strip() for line in raw_hdr.split('\r\n') if line != '']
    if hdr_lines[0] != '######## Neuralynx Data File Header':
        warnings.warn('Unexpected start to header: ' + hdr_lines[0])

    # Try to read the original file path
    try:
        try:
            assert hdr_lines[1].split()[1:3] == ['File', 'Name']
            hdr[u'FileName']  = ' '.join(hdr_lines[1].split()[3:])
            new_way = False
            # hdr['save_path'] = hdr['FileName']
        except AssertionError:
            field_name = '-OriginalFileName'
            _, value = _get_field_value(hdr_lines, field_name)
            hdr[u'FileName'] = value
            new_way = True
    except:
        warnings.warn('Unable to parse original file path from Neuralynx header: ' + hdr_lines[1])
        new_way = True

    # Process lines with file opening and closing times
    if new_way:
        parse_rest_from = 1
        time_fields = list()
        ix, hdr[u'TimeCreated'] = _get_field_value(hdr_lines, '-TimeCreated')
        time_fields.append(ix)
        hdr[u'TimeOpened_dt'] = parse_neuralynx_time_string_new(
            hdr[u'TimeCreated'])

        ix, hdr[u'TimeClosed'] = _get_field_value(hdr_lines, '-TimeClosed')
        time_fields.append(ix)
        hdr[u'TimeClosed_dt'] = parse_neuralynx_time_string_new(
            hdr[u'TimeClosed'])
    else:
        parse_rest_from = 4
        hdr[u'TimeOpened'] = hdr_lines[2][3:]
        hdr[u'TimeOpened_dt'] = parse_neuralynx_time_string(hdr_lines[2])
        hdr[u'TimeClosed'] = hdr_lines[3][3:]
        hdr[u'TimeClosed_dt'] = parse_neuralynx_time_string(hdr_lines[3])


    # Read the parameters, assuming "-PARAM_NAME PARAM_VALUE" format
    for line_idx, line in enumerate(hdr_lines[parse_rest_from:]):
        try:
            # Ignore the dash and split PARAM_NAME and PARAM_VALUE
            parts = line[1:].split()
            name = parts[0]
            value = ' '.join(parts[1:])
            if not new_way or (line_idx + parse_rest_from) not in time_fields:
                hdr[name] = value
        except:
            if not new_way or (line_idx + parse_rest_from) not in time_fields:
                warnings.warn(
                    'Unable to parse parameter line from Neuralynx header: '
                    + line)

    return hdr


def read_records(fid, record_dtype, record_skip=0, count=None):
    # Read count records (default all) from the file object fid skipping the first record_skip records. Restores the
    # position of the file object after reading.
    if count is None:
        count = -1

    pos = fid.tell()
    fid.seek(HEADER_LENGTH, 0)
    fid.seek(record_skip * record_dtype.itemsize, 1)
    rec = np.fromfile(fid, record_dtype, count=count)
    fid.seek(pos)

    return rec


def estimate_record_count(file_path, record_dtype):
    # Estimate the number of records from the file size
    file_size = os.path.getsize(file_path)
    file_size -= HEADER_LENGTH

    if file_size % record_dtype.itemsize != 0:
        warnings.warn('File size is not divisible by record size (some bytes unaccounted for)')

    return file_size / record_dtype.itemsize


def parse_neuralynx_time_string(time_string):
    # Parse a datetime object from the idiosyncratic time string in Neuralynx file headers
    try:
        tmp_date = [int(x) for x in time_string.split()[4].split('/')]
        tmp_time = [int(x) for x in time_string.split()[-1].replace('.', ':').split(':')]
        tmp_microsecond = tmp_time[3] * 1000
    except:
        warnings.warn('Unable to parse time string from Neuralynx header: ' + time_string)
        return None
    else:
        return datetime.datetime(tmp_date[2], tmp_date[0], tmp_date[1],  # Year, month, day
                                 tmp_time[0], tmp_time[1], tmp_time[2],  # Hour, minute, second
                                 tmp_microsecond)


def parse_neuralynx_time_string_new(time_string):
    # Parse a datetime object from the idiosyncratic time string in Neuralynx file headers
    try:
        if time_string == 'File was not closed properly':
            return None
        else:
            tmp_date = [int(x) for x in time_string.split()[0].split('/')]
            tmp_time = [int(x) for x in time_string.split()[-1].split(':')]
            date_list = tmp_date + tmp_time
            return datetime.datetime(*date_list)
    except:
        warnings.warn('Unable to parse time string from Neuralynx header: ' + time_string)
        return None


def check_ncs_records(records):
    # Check that all the records in the array are "similar" (have the same sampling frequency etc.
    dt = np.diff(records['TimeStamp'])
    dt = np.abs(dt - dt[0])
    if not np.all(records['ChannelNumber'] == records[0]['ChannelNumber']):
        warnings.warn('Channel number changed during record sequence')
        return False
    elif not np.all(records['SampleFreq'] == records[0]['SampleFreq']):
        warnings.warn('Sampling frequency changed during record sequence')
        return False
    elif not np.all(records['NumValidSamples'] == 512):
        warnings.warn('Invalid samples in one or more records')
        return False
    elif not np.all(dt <= 1):
        warnings.warn('Time stamp difference tolerance exceeded')
        return False
    else:
        return True


def load_ncs(file_path, load_time=True, rescale_data=True, signal_scaling=MICROVOLT_SCALING):
    # Load the given file as a Neuralynx .ncs continuous acquisition file and extract the contents
    file_path = os.path.abspath(file_path)
    with open(file_path, 'rb') as fid:
        raw_header = read_raw_header(fid)
        records = read_records(fid, NCS_RECORD)

    header = parse_header(raw_header)
    check_ncs_records(records)

    # Reshape (and rescale, if requested) the data into a 1D array
    data = records['Samples'].ravel()
    #data = records['Samples'].reshape((NCS_SAMPLES_PER_RECORD * len(records), 1))
    if rescale_data:
        try:
            # ADBitVolts specifies the conversion factor between the ADC counts and volts
            data = data.astype(np.float64) * (np.float64(header['ADBitVolts']) * signal_scaling[0])
        except KeyError:
            warnings.warn('Unable to rescale data, no ADBitVolts value specified in header')
            rescale_data = False

    # Pack the extracted data in a dictionary that is passed out of the function
    ncs = dict()
    ncs['file_path'] = file_path
    ncs['raw_header'] = raw_header
    ncs['header'] = header
    ncs['data'] = data
    ncs['data_units'] = signal_scaling[1] if rescale_data else 'ADC counts'
    ncs['sampling_rate'] = records['SampleFreq'][0]
    ncs['channel_number'] = records['ChannelNumber'][0]
    ncs['timestamp'] = records['TimeStamp']

    # Calculate the sample time points (if needed)
    if load_time:
        num_samples = data.shape[0]
        times = np.interp(np.arange(num_samples), np.arange(0, num_samples, 512), records['TimeStamp']).astype(np.uint64)
        ncs['time'] = times
        ncs['time_units'] = u'µs'

    return ncs


def load_nev(file_path):
    # Load the given file as a Neuralynx .nev event file and extract the contents
    file_path = os.path.abspath(file_path)
    with open(file_path, 'rb') as fid:
        raw_header = read_raw_header(fid)
        records = read_records(fid, NEV_RECORD)

    header = parse_header(raw_header)

    # Check for the packet data size, which should be two. DISABLED because these seem to be set to 0 in our files.
    #assert np.all(record['pkt_data_size'] == 2), 'Some packets have invalid data size'


    # Pack the extracted data in a dictionary that is passed out of the function
    nev = dict()
    nev['file_path'] = file_path
    nev['raw_header'] = raw_header
    nev['header'] = header
    nev['records'] = records
    nev['events'] = records[['pkt_id', 'TimeStamp', 'event_id', 'ttl', 'Extra', 'EventString']]

    return nev


# -*- coding: utf-8 -*-
"""
Created on Tue Feb  7 17:03:17 2023

@author: mmagnuski
"""


# %%
import os
import os.path as op

anat_dir = r'C:\Users\mmagnuski\OneDrive - Nencki PAN\RESEARCH\anat'
source_dir = op.join(anat_dir, 'sourcedata')
scan_to_postfix = {'mri': 'T1w', 'ct': 'ct'}

subjects = ['sub-U11', 'sub-U12', 'sub-U13']

for subj in subjects:
    # make sure output directory exists
    out_dir = op.join(anat_dir, subj)

    if not op.exists(out_dir):
        os.mkdir(out_dir)

    for this_scan in ['mri', 'ct']:
        this_postfix = scan_to_postfix[this_scan]
        subj_source_dir = op.join(source_dir, subj, this_scan)
        command = f'dcm2niix -o "{out_dir}" -f {subj}_{this_postfix} "{subj_source_dir}"'

        # run the command
        os.system(command)


# %% UTILS

# check n files per folder for Wrocław anat data (folder structure sometimes is messy):
data_dir = r'C:\Users\mmagnuski\OneDrive - Nencki PAN\RESEARCH\anat\sourcedata\Wrocław'

# check subfolders that have more than 100 files
def find_folders_with_many_files(data_dir, min_files=100):
    output = dict()
    for root, dirs, files in os.walk(data_dir, topdown=False):
        if len(files) >= min_files:
            output[root] = len(files)

    return output



import numpy as np


# TODO: move to borsar
# CHANGE THE api - the first argument could be a list of arrays or an xarray
def permutation_test(*arrays, paired=False, n_perm=1000, progress=False,
                     return_pvalue=True, return_distribution=True, n_jobs=1):
    import borsar

    n_groups = len(arrays)
    tail = 'both' if n_groups == 2 else 'pos'
    stat_fun = borsar.stats._find_stat_fun(n_groups=n_groups, paired=paired,
                                            tail=tail)

    thresh, dist = borsar.stats._compute_threshold_via_permutations(
        arrays, paired=paired, tail=tail, stat_fun=stat_fun,
        return_distribution=True, n_permutations=n_perm, progress=progress,
        n_jobs=n_jobs)

    stat = stat_fun(*arrays)

    # this does not make sense for > 1d,
    # maybe it didn't make sense for 1d too?
    # if isinstance(stat, np.ndarray):
    #     try:
    #         stat = stat[0]
    #     except IndexError:
    #         stat = stat.item()

    if return_pvalue:
        multiply_p = 2 if tail == 'both' else 1

        if not isinstance(stat, np.ndarray):
            if tail == 'pos' or tail == 'both' and stat >= 0:
                pval = (dist >= stat).mean() * multiply_p
            elif tail == 'neg' or tail == 'both' and stat < 0:
                pval = (dist <= stat).mean() * multiply_p

            if pval > 1.:
                pval = 1.
        else:
            if tail == 'pos':
                pval = (dist >= stat[None, :]).mean(axis=0)
            elif tail == 'neg':
                pval = (dist <= stat[None, :]).mean(axis=0)
            elif tail == 'both':
                is_pos = stat >= 0
                pval = np.zeros(stat.shape)
                pval[is_pos] = ((dist[:, is_pos] >= stat[None, is_pos])
                                .mean(axis=0))
                pval[~is_pos] = ((dist[:, ~is_pos] <= stat[None, ~is_pos])
                                .mean(axis=0))
                pval *= multiply_p

                above_one = pval > 1.
                if above_one.any():
                    pval[above_one] = 1.

    if return_distribution:
        out = dict()
        out['stat'] = stat
        out['thresh'] = thresh
        out['dist'] = dist

        if return_pvalue:
            out['pval'] = pval

        return out
    else:
        if return_pvalue:
            return stat, pval
        else:
            return stat


# TODO: auto-infer paired from xarray
def cluster_based_test(frate, compare='image', cluster_entry_pval=0.05,
                       paired=False, stat_fun=None, n_permutations=1_000,
                       n_stat_permutations=0, tail=None, progress=True):
    '''Perform cluster-based tests on firing rate data.

    Performs cluster-based test (ANOVA or t test, depending on the data) on
    firing rate to test, for example, category-selectivity of the neurons.

    Parameters
    ----------
    frate : xarray.DataArray
        Xarray with spike rate or spike density containing observations as the
        first dimension (for example trials for between-trials analysis or
        cells for between-cells analysis). If you have both cells and trials
        then you should either:
            * selected one cell at a time, via ``frate.isel(cell=0)`` for
              example
            * the trials dimension should be averaged (for example
              ``frate.mean(dim='trial')``) for a between-cell analysis.
    compare : str
        Dimension labels specified for ``'trial'`` dimension that constitutes
        categories to test selectivity for.
    cluster_entry_pval : float
        p value used as a cluster-entry threshold. The default is ``0.05``.
    paired : bool
        Whether a paired (repeated measures) or unpaired test should be used.

    Returns
    -------
    stats : numpy.ndarray
        Anova F statistics for every timepoint.
    clusters : list of numpy.ndarray
        List of cluster memberships.
    pval : numpy.ndarray
        List of p values from anova.
    '''
    from borsar.cluster import permutation_cluster_test_array

    # TODO: check if theres is a condition dimension (if so -> paired)
    arrays = [arr.values for _, arr in frate.groupby(compare)]

    if tail is None:
        n_groups = len(arrays)
        tail = 'both' if n_groups == 2 else 'pos'

    stat, clusters, pval = permutation_cluster_test_array(
        arrays, adjacency=None, stat_fun=stat_fun, threshold=None,
        p_threshold=cluster_entry_pval, paired=paired, tail=tail,
        n_permutations=n_permutations, n_stat_permutations=n_stat_permutations,
        progress=progress)

    return stat, clusters, pval


# ENH: move to sarna/borsar sometime
# ENH: allow for standard arrays (and perm_index)
def cluster_based_test_from_permutations(data, perm_data, tail='both',
                                         adjacency=None):
    '''Performs a cluster-based test from precalculated permutations.

    This function should get data ready for cluster-based permutation test
    WITHOUT the need to perform any statistical tests - so if t test between
    boots of two regions are done for example, then this should be done
    OUTSIDE of this function (the function should receive the t values in
    such case), also boot averages should be done OUTSIDE.

    Parameters
    ----------
    data : xarray.DataArray
        Data array with observations as the first dimension (for example
        trials for between-trials analysis or cells for between-cells
        analysis).
    perm_data : xarray.DataArray
        Data array with permutations. Should contain a dimension named
        ``'perm'``.
    tail : str
        Which tail to use for the test. Can be ``'both'``, ``'pos'`` or
        ``'neg'``.
    adjacency : numpy.ndarray
        Adjacency matrix for clustering. If ``None`` then lattice adjacency is
        used.

    Returns
    -------
    clusters : list of numpy.ndarray
        List of cluster memberships.
    cluster_stats : numpy.ndarray
        Cluster statistics.
    cluster_pval : numpy.ndarray
        Cluster p values.
    '''
    import xarray as xr
    import borsar

    assert isinstance(data, xr.DataArray)
    assert isinstance(perm_data, xr.DataArray)
    assert tail in ['both', 'pos', 'neg']
    dim_names = ['perm', 'permutation']
    has_dim = [dimname in perm_data.dims for dimname in dim_names]
    assert any(has_dim)
    perm_dim_name = dim_names[np.where(has_dim)[0][0]]

    percentiles = [97.5, 2.5] if tail == 'both' else [95, 5]
    perm_dim = perm_data.dims.index(perm_dim_name)
    n_perms = perm_data.shape[perm_dim]
    thresholds = np.percentile(perm_data, percentiles, axis=perm_dim)

    if tail == 'both':
        thresholds = [thresholds[0], thresholds[1]]
    elif tail == 'pos':
        thresholds = [thresholds[0], None]
    elif tail == 'neg':
        thresholds = [None, thresholds[1]]

    # clusters on actual data
    clusters, cluster_stats = borsar.find_clusters(
        data.values, thresholds, backend='borsar', adjacency=adjacency)

    # check which are pos and which neg:
    is_neg = cluster_stats < 0
    cluster_distr_pos = list()
    cluster_distr_neg = list()

    for perm_idx in perm_data.coords[perm_dim_name].data:
        perm_stat = perm_data.sel({perm_dim_name: perm_idx})

        perm_clusters, perm_cluster_stats = borsar.find_clusters(
            perm_stat.values, thresholds, backend='borsar',
            adjacency=adjacency)

        n_perm_clusters = len(perm_clusters)
        perm_is_neg = perm_cluster_stats < 0

        if n_perm_clusters > 0 and perm_is_neg.any():
            best_neg = perm_cluster_stats.min()
        else:
            best_neg = 0
        cluster_distr_neg.append(best_neg)

        if  n_perm_clusters > 0 and (~perm_is_neg).any():
            best_pos = perm_cluster_stats.max()
        else:
            best_pos = 0
        cluster_distr_pos.append(best_pos)

    cluster_distr_pos = np.array(cluster_distr_pos)
    cluster_distr_neg = np.array(cluster_distr_neg)

    cluster_pval = [(cluster_distr_pos >= clst_stat).mean() if clst_stat > 0
                    else (cluster_distr_neg <= clst_stat).mean()
                    for clst_stat in cluster_stats]
    cluster_pval = np.array(cluster_pval)

    if tail == 'both':
        # we are essentially performing two test (one for pos
        # and one for neg clusters) so we have to correct...
        cluster_pval = np.minimum(cluster_pval * 2, 1)

    return clusters, cluster_stats, cluster_pval



import numpy as np


def test_monotonic_unique_counts():
    from pylabianca._numba import _monotonic_unique_counts

    values = np.array([2, 2, 2, 5, 5, 5, 5, 5, 8, 8,
                    9, 9, 9, 9, 9, 9, 9, 9, 10, 10,
                    10, 10])
    out = _monotonic_unique_counts(values)

    assert (out[0] == np.array([ 2,  5,  8,  9, 10], dtype='int64')).all()
    assert (out[1] == np.array([3, 5, 2, 8, 4], dtype='int64')).all()



def _compute_spike_rate_numpy2(times, trials, limits, n_trials, winlen=0.25,
                               step=0.05):
    halfwin = winlen / 2
    epoch_len = limits[1] - limits[0]
    n_steps = int(np.floor((epoch_len - winlen) / step + 1))

    fr_tstart = limits[0] + halfwin
    fr_tend = limits[1] - halfwin + step * 0.001
    times = np.arange(fr_tstart, fr_tend, step=step)
    frate = np.zeros((n_trials, n_steps))

    n_epochs = event_times.shape[0]
    this_epo = trials[0]
    trilims = group(trials, diff=True)[:, 1]
    trilims = np.concatenate([[0], lims, len(trials)])

    for tridx in range(len(trilims) - 1):
        # find trials
        tms = times[trilims[tridx]:trilims[tridx + 1]]
        trial = trials[trilims[tridx]]

        # approach 1: compare with numpy array of window limits
        # approach 2: compare in a smarter way


def _select_spikes_twin(times, trials, tmin, tmax):
    msk = (times >= tmin) & (times < tmax)
    return times[msk], trials[msk]


def _old_times_trimming():
    # TODO: it makes more sense to change times to be between current
    #       time labels, than to asymmetrically trim the time labels?
    if (len(win_time) - 1) % 2 == 0:
        trim1 = int((len(win_time) - 1) / 2)
        trim2 = trim1
    else:
        trim1 = int((len(win_time) - 1) / 2)
        trim2 = trim1 + 1


import os
import os.path as op

import numpy as np
import matplotlib.pyplot as plt


def set_up_paths(onedrive_dir=None):
    """Prepare OneDrive BIDS anatomy paths.

    Parameters
    ----------
    onedrive_dir : str | None
        Path to OneDrive. If ``None``, will be detected automatically using
        ``sarna`` package.

    Returns
    -------
    paths : dict
        Dictionary of paths.
    """
    if onedrive_dir is None:
        try:
            import sarna
            onedrive_dir = sarna.proj.find_onedrive()
        except ImportError:
            raise RuntimeError('You either need to provide onedrive_dir, or'
                               'have sarna package to auto-locate OneDrive.')

    anat_dir = op.join(onedrive_dir, 'RESEARCH', 'anat')
    subjects_dir = op.join(anat_dir, 'derivatives', 'freesurfer')

    paths = {'subjects_dir': subjects_dir, 'anat_dir': anat_dir,
             'onedrive_dir': onedrive_dir}
    return paths


def plot_overlay(image, compare, title='', thresh=None):
    """Overlay plot for comparing MRI and CT scans."""
    import nibabel as nib

    image = nib.orientations.apply_orientation(
        np.asarray(image.dataobj), nib.orientations.axcodes2ornt(
            nib.orientations.aff2axcodes(image.affine))).astype(np.float32)
    compare = nib.orientations.apply_orientation(
        np.asarray(compare.dataobj), nib.orientations.axcodes2ornt(
            nib.orientations.aff2axcodes(compare.affine))).astype(np.float32)

    if thresh is not None:
        compare[compare < np.quantile(compare, thresh)] = np.nan

    fig, axes = plt.subplots(1, 3, figsize=(12, 4))

    if isinstance(title, str) and len(title) > 0:
        fig.suptitle(title)

    for i, ax in enumerate(axes):
        ax.imshow(np.take(image, [image.shape[i] // 2], axis=i).squeeze().T,
                  cmap='gray')
        ax.imshow(np.take(compare, [compare.shape[i] // 2],
                          axis=i).squeeze().T, cmap='gist_heat', alpha=0.5)
        ax.invert_yaxis()
        ax.axis('off')

    fig.tight_layout()
    return fig


def find_scans(subject, paths):
    """Find MRI and CT scans for given subjects in Labianca OneDrive BIDS
    structure.

    Parameters
    ----------
    subjects : str
        Subject name.
    paths : dict
        Dictionary of paths obtained via ``pylabianca.localize.set_up_paths``.

    Returns
    -------
    ct_dir, ct_file, mri_dir, mri_file
    """

    # we also need the path to the subject's CT scan:
    ct_dir = op.join(paths['anat_dir'], subject)
    ct_files = os.listdir(ct_dir)

    # filter files
    ct_files = [f for f in ct_files if f.endswith('.nii')
                or f.endswith('.nii.gz')]

    print('Found the following anatomy files:')
    print(ct_files)

    ct_files = [f for f in ct_files if '_ct' in f and 'preop' not in f]
    if len(ct_files) > 1:
        fname_lens = [len(f) for f in ct_files]
        longest = np.argmax(fname_lens)
        ct_fname = ct_files[longest]
    else:
        ct_fname = ct_files[0]
    print(f'Choosing {ct_fname} as the correct CT')
    ct_file = op.join(ct_dir, ct_fname)

    mri_dir = op.join(paths['subjects_dir'], subject, 'mri')
    mri_file = [f for f in os.listdir(mri_dir) if 'T1' in f][0]

    return ct_dir, ct_file, mri_dir, mri_file


def read_compute_ct_alignment(subject, paths, CT_orig, T1, plot=False):
    """Read CT -> MRI alignment matrix or compute it if it is not available.

    If the CT -> MRI alignment affine matrix is not available, but a
    freeview pre-alignment (generated by hand) is present in the relevant
    coreg subject subdirectory, then the prealignment is applied first
    and then the CT -> MRI affine is further optimized automatically.

    Parameters
    ----------
    subject : str
        Subject identifier.
    paths : dict
        Dictionary of paths obtained via ``pylabianca.localize.set_up_paths``.
    CT_orig : nibabel object
        Read CT scan in the original space (not resampled to MRI) to
        align to the MRI.
    T1 : nibabel object
        MRI T1 scan to align the CT to.
    plot : bool
        If True, plot CT-MRI alignment before and after applying the affine
        matrix.

    Return
    ------
    reg_affine : numpy.array
        The CT - MRI alignment affine matrix.
    """
    import mne
    import h5io
    import nibabel as nib

    # if pre-alignment was needed
    coreg_dir = op.join(paths['anat_dir'], 'derivatives', 'coreg', subject)
    ct_coreg_fname = op.join(coreg_dir, f'{subject}_ct_trans.h5')

    this_fname = f'{subject}_ct_aligned_manual.lta'
    pre_alg_fname = op.join(coreg_dir, this_fname)

    if op.exists(pre_alg_fname) and not op.exists(ct_coreg_fname):
        # read freeview pre-alignment transform
        manual_reg_affine_vox = mne.read_lta(pre_alg_fname)

        # convert from vox->vox to ras->ras
        manual_reg_affine = \
            CT_orig.affine @ np.linalg.inv(manual_reg_affine_vox) \
            @ np.linalg.inv(CT_orig.affine)

        with np.printoptions(suppress=True, precision=3):
            print('manual tranformation matrix:\n', manual_reg_affine)

        CT_aligned = mne.transforms.apply_volume_registration(
            CT_orig, T1, manual_reg_affine, cval='1%')


        # optimize further starting from pre-alignment
        reg_affine, _ = mne.transforms.compute_volume_registration(
            CT_orig, T1, pipeline=['rigid'],
            starting_affine=manual_reg_affine)

        # show transformation matrix after further optimization
        with np.printoptions(suppress=True, precision=3):
            print('tranformation matrix after further optimization:\n', reg_affine)

        # transform CT
        CT_aligned = mne.transforms.apply_volume_registration(
            CT_orig, T1, reg_affine, cval='1%')

        if plot:
            plot_overlay(T1, CT_aligned, 'Aligned CT - T1 preoptim',
                         thresh=0.95)

        # save
        h5io.write_hdf5(ct_coreg_fname, reg_affine, overwrite=True)

    # else/then - read the alignment or compute
    if op.exists(ct_coreg_fname):
        print('reading existing CT-MRI transform')
        reg_affine = h5io.read_hdf5(ct_coreg_fname)

    else:
        print('computing CT-MRI transform')
        reg_affine, _ = mne.transforms.compute_volume_registration(
            CT_orig, T1, pipeline='rigids')
        h5io.write_hdf5(ct_coreg_fname, reg_affine, overwrite=True)

    with np.printoptions(suppress=True, precision=3):
        print(reg_affine)

    return reg_affine


def read_channel_table(subject, paths):
    """Read table containing channel ranges and initial anatomy labels.

    Parameters
    ----------
    subject : str
        Subject identifier, for example ``'sub-U10'``.
    paths : dict
        Dictionary of paths obtained with ``pylabianca.localize.set_up_paths``.

    Returns
    -------
    chan_info : pandas.DataFrame
        DataFrame containing channel ranges and labels.
    """
    import pandas as pd

    ch_info_dir = op.join(paths['onedrive_dir'], 'RESEARCH', 'additional info',
                          'channels description')
    chan_info = pd.read_excel(
        op.join(ch_info_dir, 'All_channels_unified.xlsx'),
        sheet_name=subject.replace('sub-', ''))
    return chan_info


def read_anatomical_labels(subject, onedrive_dir):
    """Read anatomical labels assigned to channels based on DKT atlas.

    Parameters
    ----------
    subject : str
        Subject identifier.
    onedrive_dir : str
        Path to OneDrive.

    Return
    ------
    anat: pd.DataFrame
        Table with anatomical labels.
    """
    import pandas as pd

    anat_dir = op.join(onedrive_dir, 'RESEARCH', 'anat', 'derivatives',
                       'labels')
    fname = f'{subject}_labels-DKT.tsv'
    anat_table = pd.read_csv(op.join(anat_dir, fname), sep='\t')
    return anat_table


def is_known(val_or_str):
    if isinstance(val_or_str, str):
        return not val_or_str == '?'
    else:
        return True


def construct_info_from_channel_table(chan_info):
    """Construct mne-python Info object from channel table."""
    import mne
    ch_names = list()

    for row_ix in chan_info.index:
        current_row = chan_info.loc[row_ix]

        if (isinstance(current_row.electrode, str)
            and 'macro' in current_row.electrode):

            prefix = current_row.electrode.split('-')[0]
            known_range = (
                is_known(current_row['channel start'])
                and is_known(current_row['channel end'])
            )

            if not known_range:
                n_channels = 8 if prefix == 'BF' else 10
            else:
                n_channels = int(current_row['channel end']
                                 - current_row['channel start']) + 1

            area = current_row.area
            ch_name = f'{prefix}_{area}_'

            if prefix == 'BF':
                ch_names.append(ch_name + 'micro')

            for ch_idx in range(n_channels):
                ch_names.append(ch_name + f'{ch_idx + 1:01d}')

    info = mne.create_info(ch_names, sfreq=32_000, ch_types='seeg')
    return info


# TODO: separate function for reading channel labels from
#       positions unified
def read_create_channel_positions(subject, paths, ending='_micro'):
    import mne

    fname_base = f'{subject}_channel_positions_ieeg' + ending
    coreg_dir = op.join(paths['anat_dir'], 'derivatives', 'coreg', subject)


    info_fname = op.join(coreg_dir, fname_base + '.fif')
    info_fname2 = op.join(coreg_dir, fname_base + '_Karolina.fif')

    if op.exists(info_fname):
        print('Reading channel positions created by Mikołaj.')
        raw = mne.io.read_raw(info_fname)
        info = raw.info
    elif op.exists(info_fname2):
        print('Reading channel positions created by Karolina.')
        raw = mne.io.read_raw(info_fname2)
        info = raw.info
    else:
        # no saved positions found, creating new
        chan_info = read_channel_table(subject, paths)
        info = construct_info_from_channel_table(chan_info)
        print('No channel positions for this subject, you have to do '
              'some clicking!')

    return info


# select only specific regions micro channels from info
# regions -> str or list
def find_channels(info, regions=None, micro=True, side='both'):
    """Select relevant channel names and indices from mne Info object."""
    assert side.lower() in ['both', 'l', 'r']
    one_side = not side == 'both'
    regions = [regions] if isinstance(regions, str) else regions

    names = list()
    indices = list()
    for ch_idx, ch_name in enumerate(info.ch_names):
        ch_type, region_name, contact_type = ch_name.split('_')

        if regions is not None:
            is_region = any([region in region_name for region in regions])
        else:
            is_region = True

        if is_region and one_side:
            is_region = side.upper() in region_name
        if is_region and micro:
            is_region = 'micro' in contact_type

        if is_region:
            names.append(ch_name)
            indices.append(ch_idx)

    indices = np.array(indices)
    names = np.array(names)

    return names, indices


# TODO: make sure if mne.pick_info operates inplace or not
def pick_info(info, regions=None, micro=True, side='both'):
    """Subselect channels from mne Info object."""
    import mne

    _, idx = find_channels(info, regions=regions, micro=micro, side=side)
    info_sel = mne.pick_info(info.copy(), sel=idx)
    return info_sel


def project_channel_positions_to_voxels(T1, info, stat_map=True,
                                        stat_map_sd=3., channels='micro'):
    """Project channel positions from freesurfer to voxel coordinates."""
    import mne
    import borsar
    from scipy.stats import norm

    T_orig = T1.header.get_vox2ras_tkr()

    ch_pos = borsar.channels.get_ch_pos(info)
    ch_names = np.asarray(info.ch_names)

    # take only non-nan positions
    good_pos = ~ (np.isnan(ch_pos).any(axis=1))
    ch_pos = ch_pos[good_pos]
    ch_names = ch_names[good_pos]

    if channels == 'micro':
        # take only micro channels
        micro_ch_mask = np.array(['micro' in ch for ch in ch_names])
        ch_names = ch_names[micro_ch_mask]
        ch_pos = ch_pos[micro_ch_mask]

    # find channel positions in voxels
    ch_pos_mm = ch_pos * 1000  # meters → millimeters
    pos_vox = mne.transforms.apply_trans(np.linalg.inv(T_orig), ch_pos_mm)

    if stat_map:
        coords = np.indices(T1.shape)
        dist = np.sqrt(np.sum(
            (coords[None, :] - pos_vox[:, :, None, None, None]) ** 2,
            axis=1)
        )

        dist_gauss = norm.pdf(dist, loc=0, scale=stat_map_sd)
        dist_gauss /= dist_gauss.max(axis=(1, 2, 3), keepdims=True)
        all_distgauss = dist_gauss.sum(axis=0)

        return all_distgauss
    else:
        return ch_names, pos_vox


def autolabel_channels(montage, subject, paths):
    '''Auto-label channel positions with DKT atlas labels.
    The passed montage has to be in MRI space.'''

    # in: montage, subject, subjects_dir=None, aseg="aparc+aseg", dist=2
    from collections import OrderedDict
    from tqdm import tqdm

    from mne.channels import DigMontage
    from mne._freesurfer import read_freesurfer_lut, _get_aseg
    from mne.utils import _validate_type
    from mne.transforms import apply_trans
    from mne.surface import _voxel_neighbors, _VOXELS_MAX

    subjects_dir = paths['subjects_dir']
    _validate_type(montage, DigMontage, "montage")
    distances = np.arange(0.25, 5.5, step=0.25)

    aseg = 'aparc.DKTatlas+aseg'
    aseg, aseg_data = _get_aseg(aseg, subject, subjects_dir)

    # read freesurfer lookup table
    lut, fs_colors = read_freesurfer_lut()
    label_lut = {v: k for k, v in lut.items()}

    # assert that all the values in the aseg are in the labels
    assert all([idx in label_lut for idx in np.unique(aseg_data)])

    # get transform to surface RAS for distance units instead of voxels
    vox2ras_tkr = aseg.header.get_vox2ras_tkr()

    ch_dict = montage.get_positions()
    if ch_dict["coord_frame"] != "mri":
        raise RuntimeError(
            "Coordinate frame not supported, expected "
            '"mri", got ' + str(ch_dict["coord_frame"])
        )
    ch_coords = np.array(list(ch_dict["ch_pos"].values()))

    # convert to freesurfer voxel space
    ch_coords = apply_trans(
        np.linalg.inv(aseg.header.get_vox2ras_tkr()), ch_coords * 1000
    )

    # try various distances
    labels_dist = OrderedDict()
    found_labels = OrderedDict()

    for dist in tqdm(distances):
        for ch_name, ch_coord in zip(montage.ch_names, ch_coords):
            if ch_name not in found_labels:
                found_labels[ch_name] = list()
            if ch_name not in labels_dist:
                labels_dist[ch_name] = list()

            else:
                voxels = _voxel_neighbors(
                    ch_coord,
                    aseg_data,
                    dist=dist,
                    vox2ras_tkr=vox2ras_tkr,
                    voxels_max=_VOXELS_MAX,
                )
                label_idxs = set([aseg_data[tuple(voxel)].astype(int)
                                  for voxel in voxels])
                current_labels = [label_lut[idx] for idx in label_idxs]

                for lab in current_labels:
                    if lab not in found_labels[ch_name]:
                        found_labels[ch_name].append(lab)
                        labels_dist[ch_name].append((lab, dist))

    # now rename from DKT atlas names, to human-readable
    new_names = dict()
    for ch_name in labels_dist.keys():
        new_names[ch_name] = list()
        for (region, distance) in labels_dist[ch_name]:
            region = rename_region(region)
            if region is not None:
                new_names[ch_name].append((region, distance))

    return new_names


def parse_part(name):
    '''Parse parts from DKT atlas names.'''
    parts = ['superior', 'middle', 'inferior',
             'anterior', 'posterior', 'caudal', 'rostral',
             'lateral', 'medial', 'isthmus', 'transverse']

    for part in parts:
        if name.startswith(part):
            part_len = len(part)
            return part.capitalize(), name[part_len:]
    return name, None


def iterative_parsing(name):
    rest = name
    new_name = list()
    while rest is not None:
        part, rest = parse_part(rest)
        new_name.append(part.capitalize())
    return '-'.join(new_name)


def rename_region(region, ignore=None):
    '''Rename DKT region name to something more human-readable.'''
    ignore = ['White-Matter', 'Unknown', 'Vent'] if ignore is None else ignore
    translate_hemi = {'lh': 'Left', 'rh': 'Right'}

    if isinstance(region, np.str_):
        region = str(region)
    for ign in ignore:
        if ign in region:
            return None
    if region.startswith('ctx'):
        region_parts = region.split('-')
        hemi = translate_hemi[region_parts[1]]
        rest = region_parts[2]
        name = hemi + '-' + iterative_parsing(rest)
        return name
    else:
        return region


def construct_table_from_anatomical_labels(new_names):
    import pandas as pd

    df = pd.DataFrame(columns=['label', 'closest_anat', 'closest_distance',
                            'second_closest_anat', 'second_closest_distance'])

    # construct a table
    for ch_name, labels in new_names.items():
        _, ch_label, _ = ch_name.split('_')

        idx = df.shape[0]
        df.loc[idx, 'label'] = ch_label

        for label_idx, (anat_label, dist) in enumerate(labels):
            if label_idx > 1:
                break
            prefix = 'closest_' if label_idx == 0 else 'second_closest_'
            df.loc[idx, prefix + 'anat'] = anat_label
            df.loc[idx, prefix + 'distance'] = dist

    return df


def simplify_DKT_name(DKT_name, retain_side=False, translate=None):
    """Translate human-readable DKT atlas name to short labels.

    Translate the human-readable KDT atlas labels obtained with
    ``pylabianca.localize.rename_region`` to shorter anatomical labels
    we used before, like HIP, AMG or SMA."""
    assert retain_side == False  # retaining side currently not supported

    transl = {'Hippocampus': 'HIP', 'Caudal-Anterior-Cingulate': 'cACC',
              'Rostral-Anterior-Cingulate': 'rACC', 'Amygdala': 'AMG',
              'Posterior-Cingulate': 'PCC', 'Medial-Orbitofrontal': 'OFC',
              'Lateral-Orbitofrontal': 'OFC', 'Isthmus-Cingulate': 'PCC',
              'Superior-Frontal': 'SMA', 'Precuneus': 'RSC'}
    if translate is not None:
        transl = transl.update(translate)

    for tr_from, tr_to in transl.items():
        if tr_from in DKT_name:
            return tr_to

    # no simplifications found, retain original name
    return DKT_name


def create_info_from_pos(pos, ch_names=None, sfreq=None):
    import mne

    assert pos.ndim == 2
    assert pos.shape[1] == 3
    n_pos = pos.shape[0]
    if ch_names is None:
        ch_names = ['SEEG_{:03d}'.format(idx) for idx in range(n_pos)]

    sfreq = 32_000 if sfreq is None else sfreq
    dev_head_trans = mne.Transform(fro=1, to=4, trans=np.identity(4))
    info = mne.create_info(ch_names, sfreq=sfreq, ch_types='seeg')
    info['dev_head_t'] = dev_head_trans

    # apply positions
    # CHECK, CONSIDER: maybe constructing and applying montage is better?
    for ch_idx in range(n_pos):
        info['chs'][ch_idx]['loc'][:3] = pos[ch_idx, :]

    return info



import numpy as np
import numba
from numba.extending import overload


@numba.jit(nopython=True)
def _compute_spike_rate_numba(spike_times, spike_trials, time_limits,
                              n_trials, winlen=0.25, step=0.05):
    half_win = winlen / 2
    window_limits = np.array([-half_win, half_win])
    epoch_len = time_limits[1] - time_limits[0]
    n_steps = int(np.floor((epoch_len - winlen) / step + 1))

    fr_tstart = time_limits[0] + half_win
    fr_tend = time_limits[1] - half_win + step * 0.001
    times = np.arange(fr_tstart, fr_tend, step=step)
    frate = np.zeros((n_trials, n_steps))

    for step_idx in range(n_steps):
        winlims = times[step_idx] + window_limits
        msk = (spike_times >= winlims[0]) & (spike_times < winlims[1])
        tri = spike_trials[msk]
        intri, count = _monotonic_unique_counts(tri)
        frate[intri, step_idx] = count / winlen

    return times, frate


@numba.jit(nopython=True)
def _monotonic_unique_counts(values):
    n_val = len(values)
    if n_val == 0:
        return np.array([0], dtype='int64'), np.array([0], dtype='int64')

    uni = np.zeros(n_val, dtype='int64')
    cnt = np.zeros(n_val, dtype='int64')

    idx = 0
    val = values[0]
    uni[0] = val
    current_count = 1

    for next_val in values[1:]:
        if next_val == val:
            current_count += 1
        else:
            cnt[idx] = current_count
            idx += 1
            val = next_val
            uni[idx] = val
            current_count = 1

    cnt[idx] = current_count
    uni = uni[:idx + 1]
    cnt = cnt[:idx + 1]
    return uni, cnt


def numba_compare_times(spk, cell_idx1, cell_idx2, spk2=None):
    times1 = (spk.timestamps[cell_idx1] / spk.sfreq).astype('float64')

    use_spk = spk if spk2 is None else spk2
    times2 = (use_spk.timestamps[cell_idx2] / spk.sfreq).astype('float64')
    distances = np.zeros(len(times1), dtype='float64')

    res = _numba_compare_times(times1, times2, distances)
    return res


@numba.jit(nopython=True)
def _numba_compare_times(times1, times2, distances):
    n_times1 = times1.shape[0]
    n_times2 = times2.shape[0]

    tm2_idx = 0
    for idx1 in range(n_times1):
        time = times1[idx1]
        min_distance = times2.max()
        for idx2 in range(tm2_idx, n_times2):
            this_distance = np.abs(time - times2[idx2])
            if this_distance < min_distance:
                min_distance = this_distance
            else:
                break
        distances[idx1] = min_distance
        tm2_idx = max(idx2 - 1, 0)
    return distances



from typing import Type
import numpy as np
import pandas as pd

from .utils import (_deal_with_picks, _turn_spike_rate_to_xarray,
                    _get_trial_boundaries)
from .spike_rate import compute_spike_rate, _spike_density, _add_frate_info
from .spike_distance import compare_spike_times, xcorr_hist_trials


# TODO:
# - [ ] index by trial?
# - [ ] len(spk_epochs) currently gives number of units, not epochs ...
class SpikeEpochs():
    def __init__(self, time, trial, time_limits=None, n_trials=None,
                 waveform=None, waveform_time=None, cell_names=None,
                 metadata=None, cellinfo=None):
        '''Create ``SpikeEpochs`` object for convenient storage, analysis and
        visualization of spikes data.

        Parameters
        ----------
        time : listlike of np.ndarray
            List of arrays, where each array contains spike times for one
            cell (neuron). The times are centered with respect to epoch onset
            (for example stimulus presentation).
        trial : listlike of np.ndarray
            List of arrays, where each array contains trial membership of
            spikes from respective ``time`` array for one cell (neuron). The
            trial indices are zero-based integers.
        time_limits : listlike | None
            Optional. Two-element array with epoch time limits with respect to
            epoch-centering event. The limits have to be in seconds.
            For example ``np.array([-0.5, 1.5])`` means from 0.5 seconds
            before the event up to 1.5 seconds after the event. The default
            (``None``) infers time limits from min and max spike times.
        n_trials : int | None
            Number of trials. Optional, if the number of trials can't be
            inferred from the ``trials`` argument (for example when none of the
            cells fire for the last few trials).
        waveform : list of numpy ndarrays | None
            List of spikes x samples waveform arrays.
        waveform_time : np.ndarray | None
            One-dimensional array of time values in milliseconds for
            consecutive samples of the waveform.
        cell_names : list of str | None
            String identifiers of cells. First string corresponds to first
            cell, that is ``time[0]`` and ``trial[0]`` (and so forth).
            Optional, the default (``None``) names the first cell
            ``'cell000'``, the second cell ``'cell001'`` and so on.
        metadata : pandas.DataFrame
            DataFrame with trial-level metadata.
        cellinfo : pandas.DataFrame
            DataFrame with additional information about the cells.
        '''
        if not isinstance(time[0], np.ndarray):
            time = [np.asarray(x) for x in time]
        if not isinstance(trial[0], np.ndarray):
            trial = [np.asarray(x) for x in trial]
        self.time = time
        self.trial = trial

        if time_limits is None:
            tmin = min([min(x) for x in time])
            tmax = max([max(x) for x in time])
            time_limits = np.array([tmin, tmax])
        self.time_limits = time_limits

        if n_trials is None:
            n_trials = max(max(tri) + 1 if len(tri) > 0 else 0
                           for tri in self.trial)
        if cell_names is None:
            n_cells = len(time)
            cell_names = np.array(['cell{:03d}'.format(idx)
                                   for idx in range(n_cells)])
        else:
            cell_names = np.asarray(cell_names)

        if metadata is not None:
            assert isinstance(metadata, pd.DataFrame)
            assert metadata.shape[0] == n_trials

        n_cells = len(self.time)
        if cellinfo is not None:
            assert isinstance(cellinfo, pd.DataFrame)
            assert cellinfo.shape[0] == n_cells

        if waveform is not None:
            _check_waveforms(self.time, waveform, waveform_time)

        self.n_trials = n_trials
        self.waveform = waveform
        self.waveform_time = waveform_time
        self.cell_names = cell_names
        self.metadata = metadata
        self.cellinfo = cellinfo

    def __repr__(self):
        '''Text representation of SpikeEpochs.'''
        n_cells = len(self.time)
        avg_spikes = np.mean([len(x) for x in self.time])
        msg = '<SpikeEpochs, {} epochs, {} cells, {:.1f} spikes/cell on average>'
        return msg.format(self.n_trials, n_cells, avg_spikes)

# TODO: ability to get a shallow copy might also be useful
    def copy(self):
        '''Return a deep copy of the object.'''
        from copy import deepcopy
        return deepcopy(self)

    # TODO: consider if it is better to return number of epochs?
    def __len__(self):
        '''Return the number of neurons in SpikeEpochs.'''
        return len(self.time)

    def pick_cells(self, picks=None, query=None):
        '''Select cells by name or index. Operates in-place.

        Parameters
        ----------
        picks : int | str | listlike of int | list of str | None
            Cell names or indices to select.
        query : str | None
            Query for ``.cellinfo`` - to pick cells by their properties, not
            names or indices. Used only when ``picks`` is ``None``.
        '''
        if picks is None and query is None:
            return self

        if picks is None and query is not None:
            assert self.cellinfo is not None
            cellinfo_sel = self.cellinfo.query(query)
            picks = cellinfo_sel.index.values
        else:
            picks = _deal_with_picks(self, picks)

        self.time = [self.time[ix] for ix in picks]
        self.trial = [self.trial[ix] for ix in picks]
        self.cell_names = self.cell_names[picks].copy()
        if self.cellinfo is not None:
            self.cellinfo = self.cellinfo.loc[picks, :].reset_index(drop=True)
        if self.waveform is not None:
            self.waveform = [self.waveform[ix] for ix in picks]

        return self

    def drop_cells(self, picks):
        '''Drop cells by index. Operates in-place.

        Parameters
        ----------
        picks : int | str | listlike of int
            Cell  indices to drop.
        '''
        all_idx = np.arange(len(self))
        is_dropped = np.in1d(all_idx, picks)
        retain_idx = np.where(~is_dropped)[0]
        return self.pick_cells(retain_idx)

    def crop(self, tmin=None, tmax=None):
        '''Confine time range to specified limit. Operates in-place.

        Parameters
        ----------
        tmin : float | None
            Lower time bound in seconds. Spikes later or equal (``>=``) to this
            time will be retained.
        tmax : float | None
            Higher time bound in seconds. Spikes earlier or equal (``<=``) to
            this time will be retained.
        '''
        if tmin is None and tmax is None:
            raise TypeError('You have to specify tmin and/or tmax.')

        if tmin is None:
            tmin = self.time_limits[0]
        if tmax is None:
            tmax = self.time_limits[1]

        for cell_idx in range(len(self.time)):
            sel = (self.time[cell_idx] >= tmin) & (self.time[cell_idx] <= tmax)
            self.time[cell_idx] = self.time[cell_idx][sel]
            self.trial[cell_idx] = self.trial[cell_idx][sel]
        self.time_limits = [tmin, tmax]
        return self

    # TODO: refactor (DRY: merge both loops into one?)
    # TODO: better handling of numpy vs numba implementation
    # TODO: consider adding `return_type` with `Epochs` option (mne object)
    def spike_rate(self, picks=None, winlen=0.25, step=0.01, tmin=None,
                   tmax=None, backend='numpy'):
        '''Calculate spike rate with a running window.

        Parameters
        ----------
        picks : int | listlike of int | None
            The neuron index to use in the calculations. The default (``None``)
            uses all cells.
        winlen : float
            Length of the running window in seconds.
        step : float | bool
            The step size of the running window. If step is ``False`` then
            spike rate is not calculated using a running window but with
            a static one with limits defined by ``tmin`` and ``tmax``.
        tmin : float | None
            Time start in seconds. Default to trial start if ``tmin`` is ``None``.
        tmax : float | None
            Time end in seconds. Default to trial end if ``tmax`` is ``None``.
        backend : str
            Execution backend. Can be ``'numpy'`` or ``'numba'``.

        Returns
        -------
        frate : xarray.DataArray
            Xarray with following labeled dimensions: cell, trial, time.
        '''
        return compute_spike_rate(self, picks=picks, winlen=winlen, step=step,
                                  tmin=tmin, tmax=tmax, backend=backend)

    def spike_density(self, picks=None, winlen=0.3, gauss_sd=None, fwhm=None,
                      sfreq=500.):
        '''Compute spike density by convolving spikes with a gaussian kernel.

        Parameters
        ----------
        picks : int | listlike of int | None
            The neuron indices to use in the calculations. The default
            (``None``) uses all cells.
        winlen : float
            Length of the gaussian kernel window in seconds. Defaults to 0.3.
            Standard deviation of the gaussian kernel (``gauss_sd``), if
            unspecified (``gauss_sd=None``, default), is set as one sixth of
            the window length.
        gauss_sd : float | None
            Standard deviation of the gaussian kernel. By default it is set to
            ``winlen / 6``.
        fwhm : float | None
            Full width at half maximum of the gaussian kernel in seconds. If
            not ``None`` it overrides ``winlen`` and ``gauss_sd``.
        sfreq : float
            Desired sampling frequency of the spike density. Defaults to 500
            Hz.

        Returns
        -------
        frate : xarray.DataArray
            Spike density in a cell x trial x time xarray.
        '''
        picks = _deal_with_picks(self, picks)
        tms, cnt = _spike_density(self, picks=picks, winlen=winlen,
                                  gauss_sd=gauss_sd, fwhm=fwhm, sfreq=sfreq)
        xarr = _turn_spike_rate_to_xarray(
            tms, cnt.transpose((1, 0, 2)), self,
            cell_names=self.cell_names[picks])
        xarr = _add_frate_info(xarr, dep='density')

        return xarr

    def xcorr(self, picks=None, picks2=None, sfreq=500., max_lag=0.2,
              bins=None, gauss_fwhm=None):
        """
        Calculate cross-correlation histogram.

        Parameters
        ----------
        picks : ...
            List of cell indices or names to perform cross- and auto- correlations
            for. All combinations of cells will be used.
        picks2 : ...
            ...
        sfreq : float
            Sampling frequency of the bins. The bin width will be ``1 / sfreq``
            seconds. Used only when ``bins`` is ``None``. Defaults to ``500.``.
        max_lag : float
            Maximum lag in seconds. Used only when ``bins is None``. Defaults
            to ``0.2``.
        bins : numpy array | None
            Array representing edges of the histogram bins. If ``None`` (default)
            the bins are constructed based on ``sfreq`` and ``max_lag``.
        gauss_fwhm : float | None
            Full-width at half maximum of the gaussian kernel to convolve the
            cross-correlation histograms with. Defaults to ``None`` which ommits
            convolution.

        Returns
        -------
        xcorr : xarray.DataArray
            ...
        """
        return xcorr_hist_trials(
            self, picks=picks, picks2=picks2, sfreq=sfreq, max_lag=max_lag,
              bins=bins, gauss_fwhm=gauss_fwhm
        )

    def n_spikes(self, per_epoch=False):
        """Calculate number of spikes per cell (per epoch).

        Parameters
        ----------
        per_epoch: bool
            Whether to calculate number of spikes per cell splitting between
            epochs. If ``True`` the output is a ``cell x epochs`` xarray with
            additional trial info from ``.metadata`` attribute. Defaults to
            ``False``.

        Returns
        -------
        n_spikes : numpy.array | xarray.DataArray
            Number of spikes per cell. If ``per_epoch=False`` the output is
            a numpy array of spike numbers for consecutive cells. If
            ``per_epoch`` is ``True`` the output is a ``cell x epochs`` xarray
            with additional trial info from ``.metadata`` attribute.
        """
        return _n_spikes(self, per_epoch=per_epoch)

    # TODO:
    # - [ ] use `group` from sarna in looping through trials
    #       for faster execution...
    def to_neo(self, cell_idx, join=False, sep_time=0.):
        '''Turn spikes of given cell into neo.SpikeTrain format.

        Parameters
        ----------
        cell_idx : int
            Index of the cell to turn into neo.SpikeTrain format.
        join : bool | str
            Whether and how to join all the trials into a single
            neo.SpikeTrain. Defaults to ``False`` which does not perform
            any joining. If ``join=True`` or ``join='concat'``, the trials
            are concatenated into a single neo.SpikeTrain using ``sep_time``
            as time separation between consecutive trials. If ``join='pool'``,
            all the spikes are pooled and sorted with respect to their time
            (irrespective of trial number).

        Returns
        -------
        spikes : neo.SpikeTrain | list of neo.SpikeTrain
            If ``join=True``, a single neo.SpikeTrain is returned with spikes
            from separate trials concatenated (or pooled and sorted if
            ``join='pool'``. Otherwise (``join=False``) a list of
            neo.SpikeTrain objects is returned.
        '''
        import neo
        from quantities import s

        if isinstance(join, bool) and join:
            join = 'concat'

        if not join:
            spikes = list()
            trials = range(self.n_trials)
            for tri in trials:
                times = self.time[cell_idx][self.trial[cell_idx] == tri]
                spiketrain = neo.SpikeTrain(
                    times * s, t_stop=self.time_limits[1],
                    t_start=self.time_limits[0])
                spikes.append(spiketrain)
        elif join == 'concat':
            trial_len = self.time_limits[1] - self.time_limits[0]
            full_sep = trial_len + sep_time
            new_times = self.time[cell_idx] + full_sep * self.trial[cell_idx]
            t_stop = trial_len * self.n_trials + sep_time * (self.n_trials - 1)

            spikes = neo.SpikeTrain(
                new_times * s, t_stop=t_stop * s, t_start=self.time_limits[0])
        elif join == 'pool':
            times = np.sort(self.time[cell_idx])
            spikes = neo.SpikeTrain(
                times * s, t_stop=self.time_limits[1],
                t_start=self.time_limits[0])
        return spikes

    # TODO: return xarray?
    def to_raw(self, picks=None, sfreq=500.):
        '''Turn epoched spike timestamps into binned continuous representation.

        Parameters
        ----------
        spk : SpikeEpochs
            Spikes to turn into binary representation.
        picks : list-like of int | None
            Cell indices to turn to binary representation. Optional, the default
            (``None``) uses all available neurons.
        sfreq : float
            Sampling frequency of the the output array. Defaults to ``500.``.

        Returns
        -------
        times : numpy array
            1d array of time labels.
        trials_raw : numpy array
            ``trials x cells x timesamples`` array with binary spike
            information.
        '''
        return _spikes_to_raw(self, picks=picks, sfreq=sfreq)

    def __getitem__(self, selection):
        '''Select trials using an array of integers or metadata query.'''
        if isinstance(selection, str):
            if self.metadata is None:
                raise TypeError('metadata cannot be ``None`` when selecting '
                                'trials with a query.')
            # treat as pandas-style query
            new_metadata = self.metadata.query(selection)
            tri_idx = new_metadata.index.values
        elif isinstance(selection, (np.ndarray, list, tuple)):
            selection = np.asarray(selection)
            assert np.issubdtype(selection.dtype, np.integer)

            if self.metadata is not None:
                new_metadata = self.metadata.iloc[selection, :]
            tri_idx = selection
        else:
            raise TypeError('Currently only string queries are allowed to '
                            'select elements of SpikeEpochs')

        newtime, newtrial = list(), list()
        if self.metadata is not None:
            new_metadata = new_metadata.reset_index(drop=True)
        else:
            new_metadata = None

        has_waveform = self.waveform is not None
        waveform = list() if has_waveform else None

        # for each cell select relevant trials:
        for cell_idx in range(len(self.trial)):
            cell_tri = self.trial[cell_idx]
            sel = np.in1d(cell_tri, tri_idx)
            newtime.append(self.time[cell_idx][sel])

            this_tri = (cell_tri[sel, None] == tri_idx[None, :]).argmax(axis=1)
            newtrial.append(this_tri)

            if has_waveform:
                waveform.append(self.waveform[cell_idx][sel])

        new_cellinfo = None if self.cellinfo is None else self.cellinfo.copy()
        return SpikeEpochs(newtime, newtrial, time_limits=self.time_limits,
                           n_trials=len(tri_idx),
                           cell_names=self.cell_names.copy(),
                           metadata=new_metadata, cellinfo=new_cellinfo,
                           waveform=waveform)

    def plot_waveform(self, picks=None, upsample=False, ax=None, labels=True):
        '''Plot waveform heatmap for one cell.

        Parameters
        ----------
        pick : int
            Cell index to plot waveform for.
        upsample : bool | float
            Whether to upsample the waveform (defaults to ``False``). If
            ``True`` the waveform is upsampled by a factor of three. Can also
            be a value to specify the upsampling factor.
        ax : matplotlib.Axes | None
            Axis to plot to. By default opens a new figure.
        '''
        from .viz import plot_waveform
        return plot_waveform(self, picks=picks, upsample=upsample, ax=ax,
                             labels=labels, times=self.waveform_time)


def _epoch_spikes(timestamps, event_times, tmin, tmax):
    '''Epoch spike data with respect to event timestamps.

    Helper function that epochs spikes for a single neuron.

    Parameters
    ----------
    timestamps : numpy array
        Array containing spike timestamps.
    event_times : numpy array
        Array containing event timestamps.
    tmin : float
        Lower epoch limit.
    tmax : float
        Upper epoch limit.

    Returns
    -------
    trial : numpy array
        Information about the trial that the given spike belongs to.
    time : numpy array
        Spike times with respect to event onset.
    idx : numpy array
        Indices of spikes that were retained. Depending on the epoching, some
        spikes may be duplicated.
    '''
    trial = list()
    time = list()
    idx = list()

    t_idx = 0
    n_epochs = event_times.shape[0]
    this_epo = (timestamps[t_idx] < (event_times + tmax)).argmax()

    for epo_idx in range(this_epo, n_epochs):
        # find spikes that fit within the epoch
        first_idx = (timestamps[t_idx:] > (
            event_times[epo_idx] + tmin)).argmax() + t_idx
        msk = timestamps[first_idx:] < (event_times[epo_idx] + tmax)

        # select these spikes and center wrt event time
        tms = timestamps[first_idx:][msk] - event_times[epo_idx]
        if len(tms) > 0:
            tri = np.ones(len(tms), dtype='int') * epo_idx
            trial.append(tri)
            time.append(tms)

            idx.append(np.where(msk)[0] + first_idx)
        t_idx = first_idx

    if len(trial) > 0:
        trial = np.concatenate(trial)
        time = np.concatenate(time)
        idx = np.concatenate(idx)
    else:
        trial = np.array([])
        time = np.array([])

    return trial, time, idx


# TODO: make this a method of SpikeEpochs and return xarray or mne.Epochs
def _spikes_to_raw(spk, picks=None, sfreq=500.):
    '''Turn epoched spike timestamps into binary representation.

    Parameters
    ----------
    spk : SpikeEpochs
        Spikes to turn into binary representation.
    picks : list-like of int | None
        Cell indices to turn to binary representation. Optional, the default
        (``None``) uses all available neurons.
    sfreq : float
        Sampling frequency of the the output array. Defaults to ``500.``.

    Returns
    -------
    times : numpy array
        1d array of time labels.
    trials_raw : numpy array
        ``trials x cells x timesamples`` array with binary spike information.
    '''
    picks = _deal_with_picks(spk, picks)
    sample_time = 1 / sfreq
    half_sample = sample_time / 2
    tmin, tmax = spk.time_limits
    times = np.arange(tmin, tmax + 0.01 * sample_time, step=sample_time)
    time_bins = np.concatenate(
        [times - half_sample, [times[-1] + half_sample]])

    n_cells = len(picks)
    n_times = len(times)
    trials_raw = np.zeros((spk.n_trials, n_cells, n_times), dtype='int')

    for idx, cell_idx in enumerate(picks):
        trial_boundaries, tri_num = _get_trial_boundaries(spk, cell_idx)
        for ix, this_tri in enumerate(tri_num):

            spike_times = spk.time[cell_idx][
                trial_boundaries[ix]:trial_boundaries[ix + 1]]
            trials_raw[this_tri, idx, :] = np.histogram(
                spike_times, bins=time_bins
            )[0]

    return times, trials_raw


class Spikes(object):
    def __init__(self, timestamps, sfreq, cell_names=None,
                 cellinfo=None, waveform=None, waveform_time=None):
        '''Create ``Spikes`` object for convenient storage, analysis and
        visualization of spikes data.

        Parameters
        ----------
        timestamps : listlike of np.ndarray
            List of arrays, where each array contains spike timestamps for one
            cell (neuron).
        sfreq : float
            Sampling frequency of the timestamps. For example in Neuralynx
            system one timestamp occurs once per microsecond, so the sampling
            frequency is one million (``1e6``).
        cell_names : list of str | None
            String identifiers of cells. First string corresponds to first
            cell, that is ``time[0]`` and ``trial[0]`` (and so forth).
            Optional, the default (``None``) names the first cell
            ``'cell000'``, the second cell ``'cell001'`` and so on.
        cellinfo : pandas.DataFrame | None
            Additional cell information.
        waveform : list of np.ndarray
            List of spikes x samples waveform arrays.
        waveform_time : np.ndarray | None
            One-dimensional array of time values in milliseconds for
            consecutive samples of the waveform.
        '''
        n_cells = len(timestamps)
        self.timestamps = timestamps
        self.sfreq = sfreq

        if cell_names is None:
            cell_names = np.array(['cell{:03d}'.format(idx)
                                   for idx in range(n_cells)])

        self.cell_names = cell_names
        self.cellinfo = cellinfo

        if waveform is not None:
            _check_waveforms(timestamps, waveform, waveform_time)
            self.waveform = waveform
            self.waveform_time = waveform_time
        else:
            self.waveform = None
            self.waveform_time = None

    def __repr__(self):
        '''Text representation of SpikeEpochs.'''
        n_cells = len(self.cell_names)
        avg_spikes = np.mean([len(x) for x in self.timestamps])
        msg = '<Spikes, {} cells, {:.1f} spikes/cell on average>'
        return msg.format(n_cells, avg_spikes)

    # TODO: return idx from _epoch_spikes only when self.waveform is not None
    # TODO: time and consider speeding up
    def epoch(self, events, event_id=None, tmin=-0.2, tmax=1.):
        '''Epoch spikes with respect to selected events.

        Parameters
        ----------
        events : numpy.ndarray
            ``n_events x 2`` or ``n_events x 3`` array of event timestamps.
            The first column should contain event timestamp (sample) and the
            second - event type (integer).
        event_id : None | int | list=like of int
            Event types to use in epoching. The default, ``None``, uses all
            events from ``events`` array in epoching.
        tmin : float
            Epoch start in seconds with respect to event onset. Default to
            ``-0.2``.
        tmax : float
            Epoch start in seconds with respect to event onset. Default to
            ``-0.2``.

        Returns
        -------
        spk : SpikeEpochs
            Epoched spikes.
        '''
        # event_id support
        if event_id is not None:
            use_events = np.in1d(events[:, -1], event_id)
            events = events[use_events, :]

        n_neurons = len(self.timestamps)
        trial, time = list(), list()
        event_times = events[:, 0] / self.sfreq

        has_waveform = self.waveform is not None
        waveforms = list() if has_waveform else None

        for neuron_idx in range(n_neurons):
            tri, tim, idx = _epoch_spikes(
                self.timestamps[neuron_idx] / self.sfreq, event_times,
                tmin, tmax)
            trial.append(tri)
            time.append(tim)

            if has_waveform:
                waveforms.append(self.waveform[neuron_idx][idx, :])

        spk = SpikeEpochs(time, trial, time_limits=[tmin, tmax],
                          cell_names=self.cell_names, cellinfo=self.cellinfo,
                          n_trials=len(events), waveform=waveforms,
                          waveform_time=self.waveform_time)

        return spk

    # TODO: ability to get a shallow copy might also be useful
    # TODO: refactor with `.copy()` in SpikeEpochs
    def copy(self):
        '''Return a deep copy of the object.'''
        from copy import deepcopy
        return deepcopy(self)

    def __len__(self):
        '''Return the number of neurons in SpikeEpochs.'''
        return len(self.timestamps)

    # TODO: refactor out common parts with SpikeEpochs.pick_cells
    def pick_cells(self, picks=None, query=None):
        '''Select cells by name or index. Operates in-place.

        Parameters
        ----------
        picks : int | str | listlike of int | list of str | None
            Cell names or indices to select.
        query : str | None
            Query for ``.cellinfo`` - to pick cells by their properties, not
            names or indices. Used only when ``picks`` is ``None``.

        Returns
        -------
        spk : Spikes
            Selected units in a Spikes object.
        '''
        if picks is None and query is None:
            return self

        if picks is None and query is not None:
            assert self.cellinfo is not None
            cellinfo_sel = self.cellinfo.query(query)
            picks = cellinfo_sel.index.values
        else:
            picks = _deal_with_picks(self, picks)

        self.timestamps = [self.timestamps[ix] for ix in picks]
        self.cell_names = self.cell_names[picks].copy()
        if self.cellinfo is not None:
            self.cellinfo = self.cellinfo.loc[picks, :].reset_index(drop=True)
        if self.waveform is not None:
            self.waveform = [self.waveform[ix] for ix in picks]

        return self

    # TODO: DRY with SpikeEpochs
    def drop_cells(self, picks):
        '''Drop cells by index. Operates in-place.

        Parameters
        ----------
        picks : int | str | listlike of int
            Cell  indices to drop.
        '''
        all_idx = np.arange(len(self))
        is_dropped = np.in1d(all_idx, picks)
        retain_idx = np.where(~is_dropped)[0]
        return self.pick_cells(retain_idx)

    def n_spikes(self):
        """Calculate number of spikes per cell.

        Returns
        -------
        n_spikes : numpy.array
            Number of spikes per cell.
        """
        return _n_spikes(self)

    def sort(self, by=None):
        '''Sort cells. Operates in-place.

        The units are by default sorted by channel and cluster id information
        contained in dataframe stored in ``.cellinfo`` attribute.

        Parameters
        ----------
        by : str | list of str | None
            If ``None`` (default) the units are sorted by channel and cluster
            information contained in ``.cellinfo``.
            If string or list of strings - name/names of ``.cellinfo`` columns
            to sort by.
            Defaults to ``None``.

        Returns
        -------
        spk : Spikes
            Sorted Spikes.
        '''
        self = _sort_spikes(self, by)
        return self

    def plot_waveform(self, picks=None, upsample=False, ax=None, labels=True):
        '''Plot waveform heatmap for one cell.

        Parameters
        ----------
        pick : int
            Cell index to plot waveform for.
        upsample : bool | float
            Whether to upsample the waveform (defaults to ``False``). If
            ``True`` the waveform is upsampled by a factor of three. Can also
            be a value to specify the upsampling factor.
        ax : matplotlib.Axes | None
            Axis to plot to. By default opens a new figure.

        Returns
        -------
        ax : matplotlib.Axes
            Axis with waveform heatmap.
        '''
        from .viz import plot_waveform
        return plot_waveform(self, picks=picks, upsample=upsample, ax=ax,
                             labels=labels, times=self.waveform_time)

    def to_epochs(self, pad_timestamps=10_000):
        '''Turn Spike object into one-epoch SpikeEpochs representation.'''
        min_stamp = (int(min([min(x) for x in self.timestamps]))
                     - pad_timestamps)
        max_stamp = (int(max([max(x) for x in self.timestamps]))
                     + pad_timestamps)
        stamp_diff = max_stamp - min_stamp
        s_len = stamp_diff / self.sfreq

        events_fake = np.array([[min_stamp, 0, 123]])
        tmin, tmax = 0, s_len
        spk_epochs = self.epoch(events_fake, event_id=123,
                                tmin=tmin, tmax=tmax)
        return spk_epochs

    def merge(self, picks):
        '''Merge spikes from multiple cells into one. Operates in-place.

        Parameters
        ----------
        picks : list of int
            Indices of cells to merge.

        Returns
        -------
        spk : Spikes
            Modified Spikes object. The return values is to allow for chaining,
            as ``spk.merge(picks)`` operates in-place.
        '''

        picks = _deal_with_picks(self, picks)
        picks = list(picks)
        picks.sort()
        picks = picks[::-1]

        # pop timestamps in reverse order (so that indices are valid) and add
        # pop also ch_names and waveform if present
        for idx in picks[:-1]:
            if isinstance(self.cell_names, list):
                self.cell_names.pop(idx)
            else:
                self.cell_names = np.delete(self.cell_names, idx, axis=0)

            stamps = self.timestamps.pop(idx)
            self.timestamps[picks[-1]] = np.concatenate(
                (self.timestamps[picks[-1]], stamps))

            if self.waveform is not None:
                waveform = self.waveform.pop(idx)
                self.waveform[picks[-1]] = np.concatenate(
                    (self.waveform[picks[-1]], waveform))

        # sort spikes and waveforms
        ordering = np.argsort(self.timestamps[picks[-1]])
        self.timestamps[picks[-1]] = self.timestamps[picks[-1]][ordering]
        if self.waveform is not None:
            self.waveform[picks[-1]] = self.waveform[picks[-1]][ordering]

        if self.cellinfo is not None:
            # drop all but the lowest index from cellinfo
            # (agg info in cluster field?)
            self.cellinfo = self.cellinfo.drop(
                index=picks[:-1]).reset_index(drop=True)

        return self

    def to_matlab(self, path, format='osort_mm'):
        '''Save Spikes object to a matlab file in the desired format.'''
        assert format in ['fieldtrip', 'osort_mm'], 'Unknown format.'
        if format == 'fieldtrip':
            raise NotImplementedError('Sorry this is not implemented yet.')
        elif format == 'osort_mm':
            from .io import _save_spk_to_mm_matlab_format as write_spikes
        write_spikes(self, path)


def _check_waveforms(times, waveform, waveform_time):
    '''Safety checks for waveform data.'''
    n_waveforms = len(waveform)
    assert len(times) == n_waveforms

    ignore_waveforms = [x is None for x in waveform]
    n_spikes_times = np.array(
        [len(x) for x, ign in zip(times, ignore_waveforms) if not ign])
    n_spikes_waveform = np.array(
        [x.shape[0] for x, ign in zip(waveform, ignore_waveforms) if not ign])
    assert (n_spikes_times == n_spikes_waveform).all()

    if waveform_time is not None:
        n_times = len(waveform_time)
        check_waveforms = np.where(~np.array(ignore_waveforms))[0]
        n_times_wvfm_arr = [waveform[ix].shape[1]
                            for ix in check_waveforms]

        if n_waveforms > 1:
            msg = ('If `waveform_time` is passed, waveforms for each unit '
                   'need to have the same number of samples (second dimension'
                   ').')
            assert all([n_times_wvfm_arr[0] == x
                        for x in n_times_wvfm_arr[1:]]), msg

        if not n_times == n_times_wvfm_arr[0]:
            msg = ('Length of `waveform_times` and the second dimension of '
                   '`waveform` must be equal.')
            raise RuntimeError(msg)



def concatenate_spikes(spk_list, sort=True, relabel_cell_names=True):
    '''Concatenate list of spike objects into one.

    Parameters
    ----------
    spk_list : list of Spikes
        List of Spikes objects to concatenate.
    sort : bool | str | list of str
        If boolean: whether to sort the concatenated units. The units are then
        sorted by channel and cluster contained in ``.cellinfo``.
        If string or list of strings - name/names of ``.cellinfo`` columns to
        sort by. Defaults to ``True``.
    relabel_cell_names : bool
        Whether to relabel cell names to correspond to cell index.
        Defaults to ``True``.

    Returns
    -------
    spk : Spikes
        Concatenated spikes object.
    '''
    assert len(spk_list) > 0
    for spk in spk_list:
        assert isinstance(spk, Spikes), ('Not all elements in spk_list are '
                                         'Spikes objects.')

    if len(spk_list) == 1:
        return spk_list[0]

    spk = spk_list[0].copy()
    has_cellinfo = spk.cellinfo is not None
    has_waveform = spk.waveform is not None

    if has_cellinfo:
        cell_infos = [spk.cellinfo.copy()]

    for spk_add in spk_list[1:]:
        # cell names
        spk.cell_names = np.concatenate(
            [spk.cell_names, spk_add.cell_names])

        # timestamps
        spk.timestamps.extend(spk_add.timestamps)

        # cellinfo
        if has_cellinfo:
            cell_infos.append(spk_add.cellinfo)

        # waveform
        if has_waveform:
            spk.waveform.extend(spk_add.waveform)

        # metadata - only for SpikeEpochs
        # this attrib should not be present in Spikes

    if has_cellinfo:
        spk.cellinfo = pd.concat(cell_infos).reset_index(drop=True)

    if sort:
        spk = spk.sort() if isinstance(sort, bool) else spk.sort(by=sort)

    if relabel_cell_names:
        n_cells = len(spk)
        spk.cell_names = np.array(['cell{:03d}'.format(idx)
                                   for idx in range(n_cells)])

    return spk


def _sort_spikes(spk, by=None, inplace=True):
    '''Sort units by channel and cluster id or other columns in cellinfo.'''
    by = ['channel', 'cluster'] if by is None else by

    # the tests below were written by GitHub copilot entirely!
    if isinstance(by, str):
        by = [by]
    assert isinstance(by, list)
    assert all([isinstance(x, str) for x in by])
    assert all([x in spk.cellinfo.columns for x in by])

    if not inplace:
        spk = spk.copy()

    cellinfo_sorted = spk.cellinfo.sort_values(
        by=by, axis='index')
    cells_order = cellinfo_sorted.index.to_numpy()
    spk.pick_cells(cells_order)

    return spk


def _n_spikes(spk, per_epoch=False):
    """Calculate number of spikes."""
    if not per_epoch:
        if isinstance(spk, Spikes):
            return np.array([len(x) for x in spk.timestamps])
        elif isinstance(spk, SpikeEpochs):
            return np.array([len(x) for x in spk.time])
        else:
            raise TypeError("`spk` has to be an instance of Spikes or"
                            f" SpikeEpochs, got {type(spk)}.")

    else:
        if not isinstance(spk, SpikeEpochs):
            raise TypeError("When `per_epoch=True`, `spk` has to be an "
                            f"instance of SpikeEpochs, got {type(spk)}.")
        tmin, tmax = spk.time_limits
        winlen = tmax - tmin

        # FIX: this normalizes per second, we don't want that in n_spikes
        frate = compute_spike_rate(spk, step=False, tmin=tmin, tmax=tmax)
        n_spk = (frate.values * winlen).astype('int')
        return n_spk



# -*- coding: utf-8 -*-
"""
Created on Tue Mar 29 15:21:13 2022

@author: mmagnuski
"""
import os
import os.path as op

import scipy
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from scipy.stats import rankdata
from tqdm import tqdm
import h5io

import pylabianca as pln


# SETTINGS
# --------

# first channel
# to calculate coincidence only in 8-channel packs, you have to specify
# the first channel - in case you did not select any units from the first
# channel
# if you leave it as None, the first channel will be estimated from
# selected units, which may be incorrect
first_channel = 129

# the directory with sorting results - that is after manual curation and
# export using updateSORTINGresults_mm matlab function located in
# psy_screenning-\helpers\sorting_utils)
data_dir = (r'G:\.shortcut-targets-by-id\1XlCWYRlHP0YDbmo3p1NGIC6lN9XZ8l1'
            'O\switchorder\derivatives\sorting\sub-U05\ses-main\sub-U05_'
            'ses-main_task-switchorder_run-01_sorter-osort_norm-False')

# whether to plot figures and where to save them
save_fig = True
save_fig_dir = (r'G:\.shortcut-targets-by-id\1XlCWYRlHP0YDbmo3p1NGIC6lN9'
                'XZ8l1O\switchorder\derivatives\sel_ref\sub-U05\ses-main')

# data format - 'standard' or 'mm' (depends on how you exported the curated units)
data_format = 'mm'

# minimum coincidence threshold for coincidence cluster formation
coincidence_threshold = 0.25

# minimum number of different-channel units in a coincidence cluster to
# classify it as a reference cluster
min_ref_channels = 2

# weights
# -------
# weights used in calculating the final score (unit with best score is chosen)
# isi - % inter spike intervals < 3 ms (lower is better)
# nspikes  - number of spikes
# snr - signal to noise ratio (mean / std) at alignment point
# std - standard deviation calculated separately for each waveform sample
#       and then averaged (lower is better)
# dns - average "perceptual" waveform density (for each time sample 20 best
#       pixels from waveform density are chosen and averaged; these values
#       are then averaged across the whole waveform; the values are divided
#       by maximum density to de-bias against waveforms with little spikes
#       and thus make this value closer to the density we see in the figures)
weights = {'isi': 0.15, 'nspikes': 0.35, 'snr': 0.0, 'std': 0.0, 'dns': 0.5}

# alignment sample index - 94 should be true for all our osort files
algn_smpl = 94


# %% MAIN SCRIPT

# read the file
print('Reading files - including all waveforms...')
spk = pln.io.read_osort(data_dir, waveform=True, format=data_format)
print('done.')

# remove units with no spikes
n_spikes = spk.n_spikes()
no_spikes = n_spikes == 0
if (no_spikes).any():
    spk.pick_cells(~no_spikes)
    n_spikes = n_spikes[~no_spikes]


def turn_to_percentiles(arr):
    percentiles = np.vectorize(
        lambda x: scipy.stats.percentileofscore(arr, x))(arr)
    return percentiles / 100.


def plot_scores(spk_sel, score):
    from string import ascii_letters

    n_uni = len(spk_sel)
    letters = ascii_letters[1:n_uni + 1]
    ax = plt.figure(constrained_layout=True, figsize=(12, 6)).subplot_mosaic(
        'a' * n_uni + '\n' + letters,
         gridspec_kw={'height_ratios': [3, 1]}
    )

    # plot scores
    ax['a'].plot(score, marker='o')

    # plot best score in red
    best_idx = np.nanargmax(score)
    ax['a'].plot(best_idx, score[best_idx], marker='o', markersize=18,
                 markerfacecolor='r')

    # add waveforms at the bottom
    letters = list(letters)
    for ix in range(len(spk_sel)):
        spk_sel.plot_waveform(ix, ax=ax[letters[ix]], labels=False)
        ax[letters[ix]].set_title(f'{len(spk_sel.timestamps[ix])} spikes')
        ax[letters[ix]].set(yticks=[], xticks=[])

    ax['a'].set_ylabel('Weighted score', fontsize=14)
    return ax['a'].figure


# check channel packs
# -------------------
channels = spk.cellinfo.channel.unique()
channel_number = np.sort([int(ch[1:]) for ch in channels])

if first_channel is None:
    print('Estimating first channel from selected units.')
    first_channel = channel_number[0]
    print('Lowest channel number:', first_channel)

n_packs = int(np.ceil((channel_number[-1] - first_channel) / 8))
first_pack_channel = first_channel + np.arange(0, n_packs) * 8

# find which units belong to which pack
unit_channel = spk.cellinfo.channel.str.slice(1).astype('int')

units_in_pack = [np.where(np.in1d(unit_channel, frst + np.arange(8)))[0]
                 for frst in first_pack_channel]


# calculate measures
# ------------------
# isi, fr, snr, std, dns

print('Calculating FR, SNR, ISI, STD and DNS...')

n_cells = len(spk)
spk_epo = spk.to_epochs()

measure_names = ['nspikes', 'snr', 'isi', 'std', 'dns']
measures = {name: np.zeros(n_cells, dtype='float')
            for name in measure_names}
measures['nspikes'] = n_spikes

for ix in tqdm(range(n_cells)):
    # SNR
    waveform_ampl = np.mean(spk.waveform[ix][:, algn_smpl], axis=0)
    waveform_std = np.std(spk.waveform[ix][:, algn_smpl], axis=0)
    measures['snr'][ix] = np.abs(waveform_ampl) / waveform_std

    # ISI
    isi_val = np.diff(spk_epo.time[ix])
    prop_below_3ms = (isi_val < 0.003).mean()
    measures['isi'][ix] = prop_below_3ms

    # STD
    avg_std = np.std(spk.waveform[ix], axis=0).mean()
    measures['std'][ix] = avg_std

    # DNS
    measures['dns'][ix] = pln.viz.calculate_perceptual_waveform_density(
        spk, ix)

measures_prc = {name: turn_to_percentiles(measures[name])
                for name in measure_names}
print('Done.')


# compute coincidence
# -------------------
# or read from disk if already computed

fname = 'coincidence_per_pack.hdf5'
files = os.listdir(data_dir)
has_simil = fname in files

group_idx = 0
df_columns=['pack', 'group', 'channel', 'cluster', 'nspikes',
            'snr', 'isi', 'std', 'dns']

if not has_simil:
    print('Calculating similarity matrix, this can take a few minutes...')
    similarity_per_pack = list()

    for pack_idx in units_in_pack:
        if len(pack_idx) < 2:
            similarity_per_pack.append(None)
        else:
            this_spk = spk.copy().pick_cells(pack_idx)
            similarity = pln.spike_distance.compute_spike_coincidence_matrix(
                this_spk)
            similarity_per_pack.append(similarity)

    print('done.')

    # save to disk
    h5io.write_hdf5(op.join(data_dir, fname), similarity_per_pack,
                    overwrite=True)
else:
    similarity_per_pack = h5io.read_hdf5(op.join(data_dir, fname))


drop = np.zeros(len(spk), dtype='bool')
drop_perc = drop.copy()
df_list = list()

reordered_names = ['nspikes', 'snr', 'dns', 'isi', 'std']
weights_sel = np.array(
    [weights[name] for name in reordered_names]
)[None, :]

for pack_idx, (pack_units_idx, simil) in enumerate(
        zip(units_in_pack, similarity_per_pack)):

    if simil is None:
        continue

    spk_pack = spk.copy().pick_cells(pack_units_idx)
    suspicious_idx, clusters, counts = (
        pln.spike_distance.find_coincidence_clusters(
            simil,
            threshold=coincidence_threshold
        )
    )

    # detect likely REF clusters and select units
    # -------------------------------------------
    check_clst_idx = np.where(counts >= min_ref_channels)[0]

    for cluster_idx in check_clst_idx:
        print(f'Processing cluster {cluster_idx}...')
        cell_idx = pack_units_idx[suspicious_idx[clusters[cluster_idx]]]
        channels = spk.cellinfo.loc[cell_idx, 'channel'].unique()
        if len(channels) < min_ref_channels:
            continue

        group_idx += 1

        # get, rank and weight measures
        # -----------------------------
        measures_sel = np.stack(
            [measures[name][cell_idx]
             for name in ['nspikes', 'snr', 'dns']
             ] +
            [(1 - measures_prc[name][cell_idx])
             for name in ['isi', 'std']],
            axis=1
        )
        measures_perc_sel = np.stack(
            [measures_prc[name][cell_idx]
             for name in ['nspikes', 'snr', 'dns']
             ] +
            [(1 - measures_prc[name][cell_idx])
             for name in ['isi', 'std']]
        , axis=1)

        score = (measures_perc_sel * weights_sel).sum(axis=1)
        ranks = np.stack(
            [rankdata(measures_sel[:, ix])
             for ix in range(5)]
        , axis=1)
        score_ranks = (ranks * weights_sel).sum(axis=1)

        # save drop cells info according to percentiles and ranks
        save_idx = cell_idx.copy()
        save_idx_perc = np.delete(save_idx, score.argmax())
        save_idx_ranks = np.delete(save_idx, score_ranks.argmax())
        drop[save_idx_perc] = True
        drop_perc[save_idx_ranks] = True

        # fill in the dataframe
        # ---------------------
        this_df = (
            spk.cellinfo.loc[cell_idx, ['channel', 'cluster']]
            .reset_index(drop=True)
        )
        this_df.loc[:, 'group'] = group_idx
        this_df.loc[:, 'pack'] = pack_idx + 1

        # add measures
        for ix, measure_name in enumerate(reordered_names):
            this_df.loc[:, measure_name] = measures_sel[:, ix]

        df_list.append(this_df)

        # produce and save plots
        # ----------------------
        if save_fig:
            fname_base = f'pack_{pack_idx:02g}_cluster_{cluster_idx:02g}'
            fname = fname_base + '_01_coincid.png'
            fig = pln.spike_distance.plot_high_similarity_cluster(
                spk_pack, simil, clusters, suspicious_idx,
                cluster_idx=cluster_idx)
            fig.savefig(op.join(save_fig_dir, fname), dpi=300)
            plt.close(fig)

            spk_sel = spk.copy().pick_cells(cell_idx)
            fname = fname_base + '_02_score_percentiles.png'
            fig = plot_scores(spk_sel, score)
            fig.savefig(op.join(save_fig_dir, fname), dpi=300)
            plt.close(fig)

            fname = fname_base + '_03_score_within_cluster_ranks.png'
            fig = plot_scores(spk_sel, score_ranks)
            fig.savefig(op.join(save_fig_dir, fname), dpi=300)
            plt.close(fig)

    if save_fig:
        print('Plotting ignored clusters...')
        ignored_clst_idx = np.where(counts < min_ref_channels)[0]

        if len(ignored_clst_idx) > 0:
            ignored_dir = op.join(save_fig_dir, 'ignored_clusters')
            if not op.isdir(ignored_dir):
                os.mkdir(ignored_dir)

        for cluster_idx in ignored_clst_idx:
            fname_base = f'pack_{pack_idx:02g}_cluster_{cluster_idx:02g}'
            fname = fname_base + '_01_coincid.png'
            fig = pln.spike_distance.plot_high_similarity_cluster(
                spk_pack, simil, clusters, suspicious_idx,
                cluster_idx=cluster_idx)
            fig.savefig(op.join(save_fig_dir, 'ignored_clusters', fname),
                        dpi=300)
            plt.close(fig)

print('Saving dataframe to figures location...')
df = pd.concat(df_list).reset_index(drop=True)
df = df.loc[:, df_columns]
df.loc[:, 'nspikes'] = df.nspikes.astype('int')
df.to_csv(op.join(save_fig_dir, 'table.tsv'), sep='\t')

print('All done.')

# %%
# import h5io

# drop_save_dir = r'G:\.shortcut-targets-by-id\1XlCWYRlHP0YDbmo3p1NGIC6lN9XZ8l1O\switchorder\derivatives\sorting\sub-W02\ses-main'
# drop_fname = r'sub-W02_ses-main_task-switchorder_run-01_sorter-osort_norm-False_drop.hdf5'
# h5io.write_hdf5(op.join(drop_save_dir, drop_fname), drop)